/*
Copyright (c) 2001-2008, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.jtools;

import java.awt.Rectangle;
import java.awt.Dimension;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.Toolkit;
import javax.swing.JFrame;

/**
 * The ApplicationFrame class can be used to exit the application
 * gracefully after the last visible window is closed.
 * 
 * The <kbd>void cleanup(boolean isLast);</kbd> method must
 * be overwritten to do cleanup if necessary. The <kbd>isLast</kbd>
 * parameter indicates whether or not the application will exit
 * after closing the window.
 *
 * Inherited classes must not be a WindowListener theirselves.
 * A <kbd>removeWindowListener(this)</kbd> breaks the functionality.
 * 
 * @author krause
 *
 */
public abstract class ApplicationFrame extends JFrame
  implements WindowListener
{
  /** GUI controller for application.  */
  private GuiController guic = null;

  /** Internal counter for the number of visible windows */
  static private int numberOfInstances = 0;
  


  /**
   * Increment the counter (number of visible windows).
   * 
   * @return
   * The new counter value.
   */
  private synchronized int incrementNumberOfInstances() {
    numberOfInstances++;
    return numberOfInstances;
  }



  /**
   * Decrement the counter (number of visible windows).
   * 
   * @return
   * The new counter value.
   */
  private synchronized int decrementNumberOfInstances() {
    numberOfInstances--;
    if(numberOfInstances < 0) { numberOfInstances = 0; }
    return numberOfInstances;
  }



  /**
   * Create a new window.
   * 
   * @param s
   * The window title.
   */
  public ApplicationFrame(String s, GuiController c) {
    super(s); guic = c;
    incrementNumberOfInstances();
    addWindowListener(this);
  }



  /**
   * Convenience constructor.
   *
   */
  public ApplicationFrame() {
    this("", null);
  }



  /**
   * Get the current GUI controller
   *
   * @return
   * The current GUI controller
   */
  public GuiController getGuic() {
    return guic;
  }



  /**
   * Cleanup method, invoked when the window is closed.
   * 
   * @param isFinalWindow
   * Indicates whether or not this was the last visible
   * window.
   */
  abstract public void cleanup(boolean isFinalWindow);



  /**
   * Restore the window position after creating the window.
   */
  public void restorePosition() {
    Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
    Rectangle rv = new Rectangle();
    getBounds(rv);
    int x = (d.width - rv.width) / 2;
    int y = (d.height - rv.height) / 2;
    if(guic.getHaveWindowPosition()) {
      x = guic.getWindowPositionX();
      y = guic.getWindowPositionY();
      if((x + rv.width) > d.width) {
        x = d.width - rv.width;
      }
      if((y + rv.height) > d.height) {
        y = d.height - rv.height;
      }
    }
    if(x < 0) x = 0;
    if(y < 0) y = 0;
    setLocation(x, y);
  }



  /**
   * Callback method, invoked when a window is closed.
   *
   * @param e
   * The window event indicating that the window is closed.
   */
  public void windowClosing(WindowEvent e) {
    int numberOfOpenWindows = decrementNumberOfInstances();
    removeWindowListener(this);
    setVisible(false);
    if(numberOfOpenWindows == 0) {
      Rectangle rv = new Rectangle();
      getBounds(rv);
      guic.setWindowPosition(rv.x, rv.y);
      guic.writeProperties(false);
      cleanup(true);
      guic = null;
      rv = null;
      removeAll();
      dispose();
      System.gc();
      System.runFinalization();
      System.exit(0);
    } else {
      cleanup(false);
      guic = null;
      removeAll();
      dispose();
    }
  }

  /** Unused window event handler. */
  public void windowActivated(WindowEvent e) {}

  /** Unused window event handler. */
  public void windowClosed(WindowEvent e) {}

  /** Unused window event handler. */
  public void windowDeactivated(WindowEvent e) {}

  /** Unused window event handler. */
  public void windowDeiconified(WindowEvent e) {}

  /** Unused window event handler. */
  public void windowIconified(WindowEvent e) {}

  /** Unused window event handler. */
  public void windowOpened(WindowEvent e) {}

}



// project dklibsj
// @(#)ApplicationFrame.java 1.8 06/30/08

