/*
Copyright (c) 2001-2008, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.jtools;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Iterator;

import javax.swing.JPanel;

/**
 * An area to show a color.
 * 
 * @author krause
 *
 */
public class ColorShowArea extends JPanel
  implements MouseListener
{
  /** The default action command to send on mouse-click */
  public static final String defaultActionCommand = "Choose color";

  /** Flag to indicate whether or not the mouse is over the component */
  protected boolean mouseOverArea = false;

  /** Flag to indicate whether or not the mouse key is pressed */
  protected boolean mouseKeyPressed = false;
  
  /** The color to show. */
  protected Color c = null;

  /** The minimum size of the component */
  protected Dimension minimumSize = null;

  /** The preferred size of the component */
  protected Dimension preferredSize = null;

  /** The action command to send on mouse-clicks */
  protected String actionCommand = null;

  /** The list of all action listeners to notify on mouse-clicks */
  protected ArrayList<ActionListener> all = null;



  /**
   * Create a ColorShowArea.
   *
   */
  public ColorShowArea()
  {
    super();
    c = new Color(255, 255, 255);
    all = new ArrayList<ActionListener>();
    addMouseListener(this);
  }



  /**
   * Get the minimum size of the component.
   *
   * @return
   * The minimum size.
   */
  @Override
  public Dimension getMinimumSize()
  {
    Dimension back;
    if(minimumSize != null) {
      back = new Dimension(minimumSize);
    } else {
      back = new Dimension(16, 16);
    }
    return back;
  }



  /**
   * Get the preferred size of the component.
   *
   * @return
   * The preferred size.
   */
  @Override
  public Dimension getPreferredSize()
  {
    Dimension back;
    if(preferredSize != null) {
      back = new Dimension(preferredSize);
    } else {
      back = new Dimension(16, 16);
    }
    return back;
  }



  /**
   * Set the minimum size of the component.
   *
   * @param d
   * The minimum size.
   */
  @Override
  public void setMinimumSize(Dimension d)
  {
    super.setMinimumSize(d);
    minimumSize = d;
  }



  /**
   * Set the preferred size of the component.
   *
   * @param d
   * The preferred size.
   */
  @Override
  public void setPreferredSize(Dimension d)
  {
    super.setPreferredSize(d);
    preferredSize = d;
  }



  /**
   * Redraw the component.
   *
   * @param g
   * The graphics context to draw to.
   */
  @Override
  protected void paintComponent(Graphics g)
  {
    Graphics2D g2d = (Graphics2D)g.create();
    if(c != null) {
      int x = 0;
        int y = 0;
        int w = getWidth();
        int h = getHeight();
        if(mouseOverArea && isEnabled()) {
          g2d.setColor(Color.BLACK);
          g2d.fillRect(x, y, w, h);
          x++; y++; w = w - 2; h = h - 2;
          g2d.setColor(Color.WHITE);
          g2d.fillRect(x, y, w, h);
          x++; y++; w = w - 2; h = h - 2;
          if(mouseKeyPressed && isEnabled()) {
            g2d.setColor(Color.BLACK);
            g2d.fillRect(x, y, w, h);
            x++; y++; w = w - 2; h = h - 2;
            g2d.setColor(Color.WHITE);
            g2d.fillRect(x, y, w, h);
            x++; y++; w = w - 2; h = h - 2;
          }
        }
        g2d.setColor(c);
        g2d.fillRect(x, y, w, h);
    }
    g2d.dispose();
  }



  /**
   * Set the color.
   * 
   * @param co
   * The color to show.
   */
  public void setColor(Color co)
  {
    c = new Color(co.getRed(), co.getGreen(), co.getBlue());
  }



  /**
   * Get the color.
   * 
   * @return
   * The current color.
   */
  public Color getColor()
  {
    return new Color(c.getRed(), c.getGreen(), c.getBlue());
  }



  /**
   * Add action listener.
   * 
   * @param a
   * The action listener to add to the list of listeners.
   */
  public void addActionListener(ActionListener a)
  {
    all.add(a);
  }



  /**
   * Remove action listener.
   * 
   * @param a
   * The action listener to remove.
   */
  public void removeActionListener(ActionListener a)
  {
    all.remove(a);
  }



  /**
   * Set the action string.
   * 
   * @param s
   * The action string to use for creation of new events.
   */
  public void setActionCommand(String s)
  {
    actionCommand = new String(s);
  }



  /**
   * Perform action after mouse-click (choose new color)
   *
   * @param e
   * The mouse-click event.
   */
  public void mouseClicked(MouseEvent e)
  {
    if(isEnabled()) {
      String s = actionCommand;
      if(s == null) { s = defaultActionCommand; }
      Iterator<ActionListener> i = all.iterator();
      while(i.hasNext()) {
        ActionListener a = i.next();
        ActionEvent ae = new ActionEvent(this,ActionEvent.ACTION_PERFORMED, s);
        a.actionPerformed(ae);
      }
    }
  }



  /**
   * Handle mouse-pressed event.
   *
   * @param e
   * The mouse pressed event.
   */
  public void mousePressed(MouseEvent e) {
    mouseKeyPressed = true; repaint();
  }



  /**
   * Handle mouse-released event.
   *
   * @param e
   * The mouse released event.
   */
  public void mouseReleased(MouseEvent e) {
    mouseKeyPressed = false; repaint();
  }



  /**
   * Handle mouse-entered event.
   *
   * @param e
   * The mouse entered event.
   */
  public void mouseEntered(MouseEvent e) {
    mouseOverArea = true; repaint();
  }



  /**
   * Handle mouse-exited event.
   *
   * @param e
   * The mouse exited event.
   */
  public void mouseExited(MouseEvent e) {
    mouseOverArea = false; repaint();
  }
        
 

}



// vim: set ai sw=2 :



// project dklibsj
// @(#)ColorShowArea.java 1.7 06/23/08

