/*
Copyright (c) 2001-2008, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.jtools;

import java.awt.Rectangle;
import java.awt.Dimension;
import java.awt.Component;
import java.awt.Toolkit;

import javax.swing.JFrame;
import javax.swing.JDialog;

/**
 * Choose component positions related one to the other.
 *
 * @author krause
 */
public class ComponentPositioner
{

  /**
   * Choose value for one co-ordinate.
   *
   * @param sp
   * Start value of parent component.
   *
   * @param wp
   * Width value of parent component.
   *
   * @param wc
   * Width value of child component.
   *
   * @param wd
   * Width value of screen.
   *
   * @return
   * A start value for the child component.
   */
  static private int cp(int sp, int wp, int wc, int wd)
  {
    int back = 0;
    if((sp + wc) <= wd) {
      back = sp;
    } else {
      if((sp + wp) >= wc) {
	back = sp + wp - wc;
	// back = sp + wp - wc + 1;
      } else {
        back = sp + ((wp - wc) / 2);
      }
    }
    return back;
  }

  /**
   * Correct value for one co-ordinate
   *
   * @param x
   * The original (uncorrected) value.
   *
   * @param cw
   * Component width in this direction.
   *
   * @param dw
   * The screen width in this direction.
   *
   * @return
   * The corrected value.
   */
  static private int cor(int x, int cw, int dw)
  {
    int back = x;
    if((x + cw) > dw) {
      back = dw - cw;
    }
    if(back < 0) {
      back = 0;
    }
    return back;
  }

  /**
   * Choose position for component c relative to parent component p.
   *
   * @param p
   * Parent component.
   *
   * @param c
   * New child component to be placed.
   *
   * @param mod
   * Flag to indicate whether or not c is a modal component.
   */
  static public void choosePosition(Component p, Component c, boolean mod)
  {
    int x = -1;
    int y = -1;
    Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
    Rectangle pr = new Rectangle();
    Rectangle cr = new Rectangle();
    p.getBounds(pr);
    c.getBounds(cr);
    boolean b = false;
    if(!mod) {
      // right
      if((pr.x + pr.width + cr.width) < d.width) {
        x = pr.x + pr.width;
        y = cp(pr.y, pr.height, cr.height, d.height);
        b = true;
      }
      // left
      if(!b) {
        if(cr.width < pr.x) {
          x = pr.x - cr.width;
          y = cp(pr.y, pr.height, cr.height, d.height);
          b = true;
        }
      }
      // bottom
      if(!b) {
        if((pr.y + pr.height + cr.height) < d.height) {
          y = pr.y + pr.height;
          x = cp(pr.x, pr.width, cr.width, d.width);
          b = true;
        }
      }
      // top
      if(!b) {
        if(cr.height < pr.y) {
          y = pr.y - cr.height;
          x = cp(pr.x, pr.width, cr.width, d.width);
          b = true;
        }
      }
    }
    if(!b) {
      // above
      x = pr.x + ((pr.width - cr.width) / 2);
      y = pr.y + ((pr.height - cr.height) / 2);
    }
    x = cor(x, cr.width, d.width);
    y = cor(y, cr.height, d.height);
    c.setLocation(x, y);
  }

  /**
   * Choose modal position for component c relative to parent component p.
   *
   * @param p
   * Parent component.
   *
   * @param c
   * New modal child component to be placed.
   *
   */
  static public void chooseModalPosition(Component p, Component c)
  {
    choosePosition(p, c, true);
  }

  /**
   * Choose normal (non-modal) position for component c relative to
   * parent component p.
   *
   * @param p
   * Parent component.
   *
   * @param c
   * New non-modal child component to be placed.
   */
  static public void chooseNonModalPosition(Component p, Component c)
  {
    choosePosition(p, c, false);
  }


  /**
   * Correct position after resizing.
   *
   * @param c
   * The component for which the position must be corrected.
   */
  static public void correctPosition(Component c) {
    Dimension screensize = Toolkit.getDefaultToolkit().getScreenSize();
    Rectangle cr = new Rectangle();
    c.getBounds(cr);
    int x = cor(cr.x, cr.width, screensize.width);
    int y = cor(cr.y, cr.height, screensize.height);
    if((x != cr.x) || (y != cr.y)) {
      c.setLocation(x, y);
    }
  }

}



// project dklibsj
// @(#)ComponentPositioner.java 1.5 06/23/08

