/*
Copyright (c) 2001-2006, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.Insets;

import javax.swing.JPanel;

/**
 * A "factory" to create GridBagConstraints.
 * 
 * To have the same border between GUI elements an at the
 * outside of a GridBagLayout the insets on inner borders
 * must be half of the insets on outer borders.
 * 
 * @author krause
 *
 */
public class GbcMaker {
	private	int numberOfRows = 0;
	private	int numberOfColumns = 0;
	private int border = 0;
	private int th = 0;
	private int lh = 0;
	private int bh = 0;
	private int rh = 0;
	
	/**
	 * Construct a new GbcMaker
	 * 
	 * @param rows
	 * The number of rows in the GridBagLayout.
	 * 
	 * @param columns
	 * The number of columns in the GridBagLayout.
	 * 
	 * @param b
	 * The GUI element border in pixels.
	 */
	public GbcMaker(int rows, int columns, int b)
	{
		numberOfRows = rows;
		numberOfColumns = columns;
		border = b;
		th = lh = b / 2;
		bh = rh = b - th;
		if(th == 0) {
			if(b > 0) {
				th = lh = 1;
			}
		}
	}
	
	/**
	 * Create a new GridBagConstraints.
	 * 
	 * @param y
	 * The row in which the GUI element starts.
	 * 
	 * @param x
	 * The column in which the GUI element starts.
	 * 
	 * @param h
	 * The height of the GUI element in rows.
	 * 
	 * @param w
	 * The width of the GUI element in columns.
	 * 
	 * @return
	 * The GridBagConstraints to use for the GUI element.
	 */
	public GridBagConstraints create(int y, int x, int h, int w)
	{
		int t, l, r, b;
		GridBagConstraints back = new GridBagConstraints();
		back.gridx = x; back.gridy = y;
		back.gridwidth = w; back.gridheight = h;
		t = th; l = lh; r = rh; b = bh;
		if(y == 0) { t = border; }
		if(x == 0) { l = border; }
		if((y + h) >= numberOfRows) { b = border; }
		if((x + w) >= numberOfColumns) { r = border; }
		if(border > 0) {
			back.insets = new Insets(t, l, b, r);
		}
		back.anchor = GridBagConstraints.CENTER;
		back.fill = GridBagConstraints.NONE;
		back.weightx = back.weighty = 0.0;
		return back;
	}
	
	/**
	 * Create a dummy JPanel. A dummy panel can be used as
	 * a sink for free space.
	 * 
	 * @return
	 * A JPanel eating free space when growing the enclosing
	 * container.
	 */
	public JPanel createDummyPanel()
	{
		JPanel back = new JPanel();
		back.setMinimumSize(new Dimension(1,1));
		back.setPreferredSize(new Dimension(1,1));
		return back;
	}
	
	/**
	 * Create GridBagConstraints for a dummy panel (see above).
	 * 
	 * @param y
	 * The start row of the panel.
	 * 
	 * @param x
	 * The start column of the panel.
	 * 
	 * @param h
	 * The panel height in rows.
	 * 
	 * @param w
	 * The panel width in columns.
	 * 
	 * @return
	 * The GridBagConstraints for the dummy panel.
	 */
	public GridBagConstraints createDummyGbc(int y, int x, int h, int w)
	{
		GridBagConstraints back = new GridBagConstraints();
		back.gridy = y; back.gridx = x;
		back.gridheight = h; back.gridwidth = w;
		back.anchor = GridBagConstraints.CENTER;
		back.fill = GridBagConstraints.NONE;
		back.ipadx = 0; back.ipady = 0;
		back.weightx = 1.0;
		back.weighty = 1.0;
		return back;
	}
	
}



// project tracegui
// @(#)GbcMaker.java 1.6 06/23/08

