/*
Copyright (c) 2001-2008, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.jtools;

import java.lang.StringBuilder;
import java.lang.RuntimeException;
import java.util.prefs.Preferences;
import java.util.Enumeration;
import java.util.HashSet;
import java.awt.Font;
import java.awt.Toolkit;
import javax.swing.UIManager;
import javax.swing.UIDefaults;
import javax.swing.LookAndFeel;
import javax.swing.plaf.FontUIResource;
import javax.swing.UnsupportedLookAndFeelException;
import dirk_krause.jtools.StringTool;

/**
 * Control GUI.
 * The GUI controller controls the look-and-feel of a GUI
 * application.
 */
public class GuiController {

  /** The default application name */
  public static final String unknownApplication = "/unknown-application";

  /** Start string of preferences name */
  public static final String guiPrefs = "/gui";

  /** Application name for all applications */
  public static final String allApplications = "/all";

  /** Indicates whether or not the current window position is known. */
  private boolean	haveWindowPosition = false;

  /** x-value of the current window position */
  private int x = 0;

  /** y-value of the current window position */
  private int y = 0;

  /** Space between two components */
  private int b = 5;

  /** Index of LAF name in the list */
  private int lafIndex = 0;

  /** the current application name */
  private String myApplicationName = unknownApplication;

  /** Minimum font size (0=none) */
  private int minimumFontSize = 0;

  /** Flag for additional font replacements */
  private boolean doFontReplace = false;

  /** Old font name for font replacement */
  private String replaceOldName = "Dialog";

  /** New font name for font replacement */
  private String replaceNewName = "Sans";

  /** The strings used before initialize() was called. */
  private String usedFonts1 = "";

  /** The strings used after initialize() was called. */
  private String usedFonts2 = "";

  /** Key values for choosing the LAF */
  public static final String[] lafNamesShort = {
    /*  0 */	"system",
    /*  1 */	"none"
  };

  /** Preferences key */
  public static final String[] pk = {
    /*  0 */	"border",
    /*  1 */	"window.x",
    /*  2 */	"window.y",
    /*  3 */	"look-and-feel",
    /*  4 */	"fonts.replace",
    /*  5 */	"fonts.old-name",
    /*  6 */	"fonts.new-name",
    /*  7 */	"fonts.minmum-size",
    /*  8 */	"",
    /*  9 */	", ",
  };

  /** Flag to indicate whether initialization was not yet done */
  protected boolean notYetInitialized = true;



  /**
   * Get integer value from preferences.
   *
   * @param k
   * The preference key (name).
   *
   * @param p1
   * Application-specific preferences.
   *
   * @param p2
   * All-application preferences.
   *
   * @param v
   * Default value.
   *
   * @return
   * Value found in preferences or v.
   */
  private int getInt(
    String k, Preferences p1, Preferences p2, int v
  )
  {
    int back = v;
    int i;
    if(p2 != null) {
      try {
        i = p2.getInt(k, back);
	back = i;
      } catch(RuntimeException e) {}
    }
    if(p1 != null) {
      try {
        i = p1.getInt(k, back);
	back = i;
      } catch(RuntimeException e) {}
    }
    return back;
  }



  /**
   * Get boolean value from preferences.
   *
   * @param k
   * Preference name.
   *
   * @param p1
   * Application-specific preferences.
   *
   * @param p2
   * All-applications preferences.
   *
   * @param v
   * Default value.
   *
   * @return
   * The value found in the preferences or v.
   */
  private boolean getBoolean(
    String k, Preferences p1, Preferences p2, boolean v
  )
  {
    boolean back = v;
    boolean b;
    if(p2 != null) {
      try {
        b = p2.getBoolean(k, back);
	back = b;
      } catch(RuntimeException e) {}
    }
    if(p1 != null) {
      try {
        b = p1.getBoolean(k, back);
	back = b;
      } catch(RuntimeException e) {}
    }
    return back;
  }



  /** 
   * Get string value from preferences
   *
   * @param k
   * The preference name.
   *
   * @param p1
   * Application-specific preferences.
   *
   * @param p2
   * All-applications preferences.
   *
   * @param v
   * Default value.
   *
   * @return
   * The value found in the preferences or v.
   */
  private String getString(String k, Preferences p1, Preferences p2, String v)
  {
    String back = v;
    String s;
    if(p2 != null) {
      try {
        s = p2.get(k, back);
	back = s;
      } catch(RuntimeException e) {}
    }
    if(p1 != null) {
      try {
        s = p1.get(k, back);
	back = s;
      } catch(RuntimeException e) {}
    }
    return back;
  }




  /**
   * Save integer value to preferences.
   *
   * @param k
   * Preference name.
   *
   * @param p1
   * Application-specific preferences.
   *
   * @param p2
   * All-application preferences.
   *
   * @param v
   * The value to save.
   *
   * @param f
   * Flag to store value for all applications.
   */
  private void putInt(
    String k, Preferences p1, Preferences p2, int v, boolean f
  )
  {
    if(p1 != null) {
      try {
        p1.putInt(k, v);
      } catch(RuntimeException e) {}
    }
    if(f) {
    if(p2 != null) {
      try {
        p2.putInt(k, v);
      } catch(RuntimeException e) {}
    }
    }
  }



  /**
   * Save string value to preferences.
   *
   * @param k
   * Preference name.
   *
   * @param p1
   * Application-specific preferences.
   *
   * @param p2
   * All-application preferences.
   *
   * @param v
   * The value to save.
   *
   * @param f
   * Flag to store value for all applications.
   */
  private void putString(
    String k, Preferences p1, Preferences p2, String v, boolean f
  )
  {
    if(p1 != null) {
      try {
        p1.put(k, v);
      } catch(RuntimeException e) {}
    }
    if(f) {
    if(p2 != null) {
      try {
        p2.put(k, v);
      } catch(RuntimeException e) {}
    }
    }
  }



  /**
   * Save boolean value to preferences.
   *
   * @param k
   * Preference name.
   *
   * @param p1
   * Application-specific preferences.
   *
   * @param p2
   * All-application preferences.
   *
   * @param v
   * The value to save.
   *
   * @param f
   * Flag to store value for all applications.
   */
  private void putBoolean(
    String k, Preferences p1, Preferences p2, boolean v, boolean f
  )
  {
    if(p1 != null) {
      try {
        p1.putBoolean(k, v);
      } catch(RuntimeException e) {}
    }
    if(f) {
    if(p2 != null) {
      try {
        p2.putBoolean(k, v);
      } catch(RuntimeException e) {}
    }
    }
  }



  /**
   * Get the LAF index.
   *
   * @return
   * The LAF index.
   */
  protected synchronized int getLafIndex() {
    int back; back = lafIndex; return back;
  }



  /**
   * Set new LAF index.
   *
   * @param n
   * The new LAF index.
   */
  protected synchronized void setLafIndex(int n) {
    lafIndex = n;
  }



  /**
   * Constructor support function.
   *
   * @param n		Application name
   * @param useToolkit	Flag: Use toolkit to find font size. 
   */
  private void construct(String n, boolean useToolkit) {
    myApplicationName = n;
    Preferences pun = null; Preferences pua = null;
    try {
      pun = Preferences.userRoot().node(guiPrefs + "/" + n);
    } catch(RuntimeException e) { pun = null; }
    try {
      pua = Preferences.userRoot().node(guiPrefs + allApplications);
    } catch(RuntimeException e) { pua = null; }
    b = getInt(pk[0], pun, pua, b);
    x = getInt(pk[1], pun, pua, -1);
    y = getInt(pk[2], pun, pua, -1);
    if((x > -1) && (y > -1)) { haveWindowPosition = true; }
    String s = getString(pk[3], pun, pua, lafNamesShort[0]);
    lafIndex = StringTool.getArrayIndex(lafNamesShort, s);
    lafIndex = toRange(lafIndex, 0, (lafNamesShort.length - 1));
    doFontReplace = getBoolean(pk[4], pun, pua, false);
    replaceOldName = getString(pk[5], pun, pua, replaceOldName);
    replaceNewName = getString(pk[6], pun, pua, replaceNewName);
    if(useToolkit) {
      minimumFontSize =
      (14 * Toolkit.getDefaultToolkit().getScreenResolution()) / 98;
    }
    minimumFontSize = getInt(pk[7], pun, pua, minimumFontSize);
  }


  private void construct(String n) {
    construct(n, false);
  }




  /**
   * Write settings back to preferences.
   *
   * @param forAll
   * Flag to indicate whether or not to write preferences
   * for all applications.
   */
  public void writeProperties(boolean forAll) {
    Preferences pun = null; Preferences pua = null;
    try {
      pun = Preferences.userRoot().node(guiPrefs + "/" + myApplicationName);
    } catch(RuntimeException e) { pun = null; }
    try {
      pua = Preferences.userRoot().node(guiPrefs + allApplications);
    } catch(RuntimeException e) { pua = null; }
    putInt(pk[0], pun, pua, getBorder(), forAll);
    putInt(pk[1], pun, pua, getWindowPositionX(), false);
    putInt(pk[2], pun, pua, getWindowPositionY(), false);
    putString(pk[3], pun, pua, lafNamesShort[getLafIndex()], forAll);
    putBoolean(pk[4], pun, pua, getDoFontReplace(), forAll);
    putString(pk[5], pun, pua, getReplaceOldName(), forAll);
    putString(pk[6], pun, pua, getReplaceNewName(), forAll);
    putInt(pk[7], pun, pua, getMinimumFontSize(), forAll);
  }



  /**
   * Constructor using application name.
   *
   * @param applicationName
   * The application name.
   */
  public GuiController(String applicationName) {
    construct(applicationName);
  }

  /**
   *
   */
  public GuiController(String applicationName, boolean useToolkit)
  {
    construct(applicationName, useToolkit);
  }


  /**
   * Constructor without application name.
   *
   * The default application name is used.
   */
  public GuiController() {
    construct(unknownApplication);
  }



  /**
   * Check whether the current window position is known.
   *
   * @return
   * Boolean indicator flag.
   */
  public synchronized boolean getHaveWindowPosition() {
    return haveWindowPosition;
  }
  


  /**
   * Get current window x position.
   *
   * @return
   * The current x position.
   */
  public synchronized int getWindowPositionX() {
    return x;
  }



  /**
   * Get the current window y position.
   *
   * @return
   * The current window y position.
   */
  public synchronized int getWindowPositionY() {
    return y;
  }



  /**
   * Set the current window position
   *
   * @param newx
   * The current x position.
   *
   * @param newy
   * The current y position.
   */
  public synchronized void setWindowPosition(int newx, int newy) {
    x = ((newx > 0) ? newx : 0);
    y = ((newy > 0) ? newy : 0);;
    haveWindowPosition = true;
  }



  /**
   * Retrieve current space between components.
   *
   * @return
   * The space to leave between two components.
   */
  public synchronized int getBorder() {
    return b;
  }




  /**
   * Set new component space.
   *
   * @param bo
   * The new space value.
   */
  public synchronized void setBorder(int bo) {
    b = bo;
  }



  /**
   * Get the minimum font size.
   *
   * @return
   * The minimum font size.
   */
  public synchronized int getMinimumFontSize() {
    int b; b = minimumFontSize; return b;
  }



  /**
   * Set the minimum font size.
   *
   * @param nfs
   * The new minimum font size.
   */
  public synchronized void setMinimumFontSize(int nfs) {
    minimumFontSize = nfs;
  }



  /**
   * Check whether the LAF is already initialized.
   *
   * @return
   * Flag indicating whether LAF setup still must be done.
   */
  private synchronized boolean mustInitialize() {
    boolean back = notYetInitialized;
    notYetInitialized = false;
    return back;
  }



  /**
   * Check whether or not to replace fonts.
   *
   * @return
   * Flag indicating whether or not font replacement must be done.
   */
  protected synchronized boolean getDoFontReplace() {
    boolean back; back = doFontReplace; return back;
  }




  /**
   * Set font replacement flag.
   *
   * @param f
   * New value for flag.
   */
  protected synchronized void setDoFontReplace(boolean f) {
    doFontReplace = f;
  }



  /**
   * Get old name for font replacement.
   *
   * @return
   * Font name of font to replace.
   */
  protected synchronized String getReplaceOldName() {
    String s; s = replaceOldName; return s;
  }



  /**
   * Set old name for font replacement.
   *
   * @param s
   * Name of font to replace.
   */
  protected synchronized void setReplaceOldName(String s) {
    replaceOldName = s;
  }



  /**
   * Get new name for font replacement.
   *
   * @return
   * New font name.
   */
  protected synchronized String getReplaceNewName() {
    String s; s = replaceNewName; return s;
  }



  /**
   * Set new name for font replacement.
   *
   * @param s
   * New name of the font.
   */
  protected synchronized void setReplaceNewName(String s) {
    replaceNewName = s;
  }


  private int toRange(int v, int min, int max) {
    int back = v;
    if(back < min) { back = min; }
    if(back > max) { back = max; }
    return back;
  }



  public synchronized String getFontsUsed1() {
    String s; s = usedFonts1; return s;
  }

  public synchronized String getFontsUsed2() {
    String s; s = usedFonts2; return s;
  }

  private synchronized void setFontsUsed1(String s) {
    usedFonts1 = s;
  }

  private synchronized void setFontsUsed2(String s) {
    usedFonts2 = s;
  }



  private void fontModifications(
    UIDefaults d, HashSet<String> s1, HashSet<String> s2
  ) {
    Enumeration en = d.keys();			// defaults keys collection
    HashSet<Object> hs = new HashSet<Object>();	// objects we need to process
    boolean must_replace;			// flag: must replace
    boolean do_font_replacements = false;	// must do font replacements
    Object	k = null;			// key object
    Object	v = null;			// value object
    Font	f = null;			// font
    FontUIResource	fuir = null;		// FontUIResource
    String	font_name = null;		// font name
    int		font_style = 0;			// font style
    int		font_size = 0;			// font size
    String	old_font_name = null;		// old font name
    String	new_font_name = null;		// new font name
    int		min_font_size = 0;
    do_font_replacements = getDoFontReplace();
    if(do_font_replacements) {
      old_font_name = getReplaceOldName();
      new_font_name = getReplaceNewName();
    }
    min_font_size = getMinimumFontSize();
    while(en.hasMoreElements()) {
      try {
        k = en.nextElement();
        v = d.get(k);
	if(v != null) {
	  if((v instanceof Font) || (v instanceof FontUIResource)) {
	    hs.add(k);
	  }
	}
      } catch(Exception e) {}
      Object[] keyArray;
      keyArray = hs.toArray();
      for(int i = 0; i < keyArray.length; i++) {
        must_replace = false;
	k = keyArray[i];
	try {
	  v = d.get(k);
	  if(v != null) {
	    if(v instanceof Font) {
	      f = (Font)v;
	      font_name = f.getName();
	      font_style = f.getStyle();
	      font_size = f.getSize();
	      s1.add(
	        new String(
		  font_name + " " + (new Integer(font_size)).toString()
		)
	      );
	      if(do_font_replacements) {
	        if(font_name.equals(old_font_name)) {
		  font_name = new_font_name; must_replace = true;
		}
	      }
	      if(min_font_size > 0) {
	        if(font_size < min_font_size) {
		  font_size = min_font_size; must_replace = true;
		}
	      }
	      if(must_replace) {
	        d.put(
		  k,
		  (Object)(new Font(font_name, font_style, font_size))
		);
	      }
	      s2.add(
	        new String(
		  font_name + " " + (new Integer(font_size)).toString()
		)
	      );
	    } else {
	      if(v instanceof FontUIResource) {
	        fuir = (FontUIResource)v;
		font_name = fuir.getName();
		font_style = fuir.getStyle();
		font_size = fuir.getSize();
		s1.add(
		  new String(
		    font_name + " " + (new Integer(font_size)).toString()
		  )
		);
		if(do_font_replacements) {
		  if(font_name.equals(old_font_name)) {
		    font_name = new_font_name; must_replace = true;
		  }
		}
		if(min_font_size > 0) {
		  if(font_size < min_font_size) {
		    font_size = min_font_size; must_replace = true;
		  }
		}
		if(must_replace) {
		  d.put(
		    k,
		    (Object)(new FontUIResource(
		      font_name, font_style, font_size
		    ))
		  );
		}
		s2.add(
		  new String(
		    font_name + " " + (new Integer(font_size)).toString()
		  )
		);
	      }
	    }
	  }
	} catch(Exception e) {}
      }
    }
  }



  private String fontHashSetToString(HashSet<String> shs) {
    StringBuilder sb = new StringBuilder();
    Object k = null;
    boolean is_not_first = false;
    Object keyArray[] = shs.toArray();
    for(int i = 0; i < keyArray.length; i++) {
      k = keyArray[i];
      if(is_not_first) {
        sb.append(pk[9]);
      }
      sb.append((String)k);
      is_not_first = true;
    }
    String back = sb.toString();
    return back;
  }



  public void initialize() {
    if(mustInitialize()) {
      HashSet<String> s1 = new HashSet<String>();	// old fonts
      HashSet<String> s2 = new HashSet<String>();	// new fonts
      // step 1: set LAF and correct defaults
      switch(getLafIndex()) {
        case 0: {
	  try {
	    // String s = UIManager.getCrossPlatformLookAndFeelClassName();
	    String s = UIManager.getSystemLookAndFeelClassName();
	    UIManager.setLookAndFeel(s);
	    LookAndFeel laf = UIManager.getLookAndFeel();
	    if(laf != null) {
	      UIDefaults defs = laf.getDefaults();
	      fontModifications(defs, s1, s2);
	    }
	  } catch(Exception e) {
	  }
	} break;
      }
      // step 2: correct defaults
      UIDefaults d2 = UIManager.getDefaults();
      fontModifications(d2, s1, s2);
      // step 3: set font strings
      setFontsUsed1(fontHashSetToString(s1));
      setFontsUsed2(fontHashSetToString(s2));
    }
  }


}

// vim: set ai sw=2 :


// project dklibsj
// @(#)GuiController.java 1.10 10/27/09

