/*
Copyright (c) 2009, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package bmepsgui;

import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.File;

import java.util.ArrayList;
import java.util.TreeMap;

import dirk_krause.tools.*;


/**
 *
 * @author krause
 */
public class BmepsController {
    /** Tool. */
    private DkTool dkt = null;

    /** Gui controller. */
    private DkGuiTool dkgt = null;

    /** Flag: The bmeps program was found. */
    private boolean bmepsFound = false;

    /** Configurations in the bmeps program. */
    private String[] builtinConfigurations = null;

    /** Configurations from bmeps configuration file. */
    private String[] fileConfigurations = null;

    public static final String[] cmdToReadConfigFile = {
        "bmeps", "--configuration-file", "--/ui/lang=en",
        "--/ui/lang/env=no", "--/log/stdout/level=none",
        "--/log/stderr/level=none", "--/log/file/level=none"
    };

    public static final String[] baseDrivers = {
        "ps",
        "pdf",
        "bb",
        "eps",
    };

    public static final String[] pdfVersions = {
        "1.2", "1.3", "1.4",
    };

    public static final String[] pdfFit = {
        "width", "height", "page"
    };

    public static final String[] outputSuffixes = {
        ".eps", ".pdf", ".bb", ".ps",
    };

    public static final String[] bmepsOptions = {
        /*   0 */ "bmeps",
        /*   1 */ "-r",
        /*   2 */ "-l",
	/*   3 */ ",d=y",
	/*   4 */ ",l=",
	/*   5 */ ",dsc.c=y",
	/*   6 */ ",dsc.c=n",
	/*   7 */ ",c=y",
	/*   8 */ ",c=n",
	/*   9 */ ",i=y",
	/*  10 */ ",i=n",
	/*  11 */ ",j.i=y",
	/*  12 */ ",j.i=n",
	/*  13 */ ",e=",
	/*  14 */ ":",
	/*  15 */ "a",
	/*  16 */ "r",
	/*  17 */ "f",
	/*  18 */ ",m.d.s=y",
	/*  19 */ ",m.d.s=n",
	/*  20 */ "d",
	/*  21 */ ",f.d.p=y",
	/*  22 */ ",f.d.p=n",
	/*  23 */ ",j.s.s=",
	/*  24 */ ",u.r.c=y",
	/*  25 */ ",m.s=",
	/*  26 */ ",m=y",
	/*  27 */ ",m=n",
	/*  28 */ ",b=",
	/*  29 */ ",a.u.d.b=y",
	/*  30 */ ",a.u.d.b=n",
	/*  31 */ ",c.i.m=y",
	/*  32 */ ",c.i.m=n",
	/*  33 */ ",i.m.t.l=0.0001",
	/*  34 */ ",i.m.t.l=1",
	/*  35 */ ",s.d=y",
	/*  36 */ ",s.d=n",
	/*  37 */ ",v=y",
	/*  38 */ ",v=n",
	/*  39 */ ",o.d=y",
	/*  40 */ ",sh=y",
	/*  41 */ ",sh=n",
	/*  42 */ ",p.f=",
	/*  43 */ ",t.a.c=y",
	/*  44 */ ",t.a.c=n",
	/*  45 */ ",a.p.p.a=y",
	/*  46 */ ",a.p.p.a=n",
        /*  47 */ "--/ui/lang/env=off",
	/*  48 */ "--/ui/lang=",
	/*  49 */ ".UTF-8",
	/*  50 */ "--/log/stdout/level=info",
	/*  51 */ "--/log/stderr/level=none",
	/*  52 */ "--/log/file/level=none",
	/*  53 */ "--/log/stdout/time=yes",
	/*  54 */ "--/log/stdout/split=yes",
	/*  55 */ "--/log/file/keep=none",
	/*  56 */ "--/log/stdout/codepage=-",
        /*  57 */ ",predictor=",
        /*  58 */ "none",
        /*  59 */ "tiff",
        /*  60 */ "png-sub",
        /*  61 */ "png-up",
        /*  62 */ "png-average",
        /*  63 */ "png-paeth",
    };

    /** Preference keys to retrive/save strings. */
    private static final String[] pkStr = {
        "directory.current",    //  0: Working directory
        "config.built-in",      //  1: Built-in configuration
        "ps.size.media",        //  2: PS media size
        "pdf.size.media",
        "config.from-file",
    };

    /** Private string values. */
    protected String[] myStr = {
        null,       //  0: Working directory
        "eps2",     //  1: Built-in configuration
        "A4",       //  2: PS media size
        "A4",       //  3: PDF media size
        null,       //  4: Configuration from file
    };

    private static final String[] pkBool = {
        "make",
        "draft",
        "config.full-custom",
        "config.from-file",
        "ps.encapsulated",
        "ps.dsc",
        "ps.color",
        "ps.interpolation",
        "ps.interpolation.jpeg",
        "ps.encoding.ascii85",
        "ps.encoding.run-length",
        "ps.encoding.separated-channels",
        "ps.encoding.flate",
        "ps.encoding.dct",
        "ps.encoding.dct.ignore-jfif-analysis",
        "ps.encoding.dct.sof00",
        "ps.encoding.dct.sof01",
        "ps.encoding.dct.sof02",
        "ps.encoding.dct.sof03",
        "ps.encoding.dct.sof04",
        "ps.encoding.dct.sof05",
        "ps.encoding.dct.sof06",
        "ps.encoding.dct.sof07",
        "ps.encoding.dct.sof08",
        "ps.encoding.dct.sof09",
        "ps.encoding.dct.sof10",
        "ps.encoding.dct.sof11",
        "ps.encoding.dct.sof12",
        "ps.encoding.dct.sof13",
        "ps.encoding.dct.sof14",
        "ps.encoding.dct.sof15",
        "ps.alpha.mix",
        "ps.alpha.ingore-bg-chunk",
        "ps.alpha.image-mask",
        "ps.alpha.image-mask.require-full-opacity",
        "ps.operator.dictionary",
        "ps.operator.vmreclaim",
        "ps.operator.showpage",
        "pdf.interpolation",
        "pdf.interpolation.jpeg",
        "pdf.encoding.dct.ignore-jfif-analysis",
        "pdf.encoding.dct.sof00",
        "pdf.encoding.dct.sof01",
        "pdf.encoding.dct.sof02",
        "pdf.encoding.dct.sof03",
        "pdf.encoding.dct.sof04",
        "pdf.encoding.dct.sof05",
        "pdf.encoding.dct.sof06",
        "pdf.encoding.dct.sof07",
        "pdf.encoding.dct.sof08",
        "pdf.encoding.dct.sof09",
        "pdf.encoding.dct.sof10",
        "pdf.encoding.dct.sof11",
        "pdf.encoding.dct.sof12",
        "pdf.encoding.dct.sof13",
        "pdf.encoding.dct.sof14",
        "pdf.encoding.dct.sof15",
        "pdf.alpha.transfer",
        "pdf.alpha.attributes",
        "pdf.alpha.mix",
        "pdf.alpha.ingore-bg-chunk",
        "pdf.alpha.image-mask",
        "pdf.alpha.image-mask.require-full-opacity",
    };

    protected boolean[] myBool = {
        false,      //  0: make-mode
        false,      //  1: draft
        false,      //  2: full-custom
        false,      //  3: from configuration file
        true,       //  4: PS encapsulated
        false,      //  5: PS DSC comments
        true,       //  6: PS color
        true,       //  7: PS image interpolation
        false,      //  8: PS image interpolation for JPEGs too
        true,       //  9: PS ASCII-85
        true,       // 10: PS run-length
        true,       // 11: PS run-length multiple data channels
        false,      // 12: PS flate compression
        true,       // 13: PS DCT encoding for JPEG
        true,       // 14: PS DCT encoding ignores JFIF analysis result
        true,       // 15: PS SOF 0
        true,       // 16: PS SOF 1
        false,      // 17: PS SOF 2
        false,      // 18: PS SOF 3
        false,      // 19: PS SOF 4
        false,      // 20: PS SOF 5
        false,      // 21: PS SOF 6
        false,      // 22: PS SOF 7
        false,      // 23: PS SOF 8
        false,      // 24: PS SOF 9
        false,      // 25: PS SOF 10
        false,      // 26: PS SOF 11
        false,      // 27: PS SOF 12
        false,      // 28: PS SOF 13
        false,      // 29: PS SOF 14
        false,      // 30: PS SOF 15
        true,       // 31: PS mix alpha
        false,      // 32: PS alpha ignore bg chunk
        false,      // 33: PS alpha image mask
        false,      // 34: PS alpha image mask require full opacity
        true,       // 35: PS dictionary
        false,      // 36: PS vmreclaim
        false,      // 37: PS showpage
        true,       // 38: PDF image interpolation
        false,      // 39: PDF image interpolation for JPEGs too
        true,       // 40: PDF ignore JFIF analysis result
        true,       // 41: PS SOF 0
        true,       // 42: PS SOF 1
        false,      // 43: PS SOF 2
        false,      // 44: PS SOF 3
        false,      // 45: PS SOF 4
        false,      // 46: PS SOF 5
        false,      // 47: PS SOF 6
        false,      // 48: PS SOF 7
        false,      // 49: PS SOF 8
        false,      // 50: PS SOF 9
        false,      // 51: PS SOF 10
        false,      // 52: PS SOF 11
        false,      // 53: PS SOF 12
        false,      // 54: PS SOF 13
        false,      // 55: PS SOF 14
        false,      // 56: PS SOF 15
        true,       // 57: PDF alpha transfer
        true,       // 58: PDF alpha page attributes
        false,      // 59: PDF alpha mix
        false,      // 60: PDF alpha mix ignore bg chunk
        false,      // 61: PDF alpha image mask
        false,      // 62: PDF alpha image mask require full opacity
    };

    private static final String[] pkInt = {
        "driver.base",          //  0: Base driver
        "ps.version",           //  1: PS Version
        "ps.size",              //  2: PS size
        "ps.alpha.mix.red",
        "ps.alpha.mix.green",
        "ps.alpha.mix.blue",
        "pdf.version",
        "pdf.fit",
        "pdf.size",
        "pdf.alpha.mix.red",
        "pdf.alpha.mix.green",
        "pdf.alpha.mix.blue",
        "dialog.tab.ps",
        "dialog.tab.pdf",
        "pdf.predictor",
    };

    protected int[] myInt = {
        0,          //  0: Base driver (0=PS, 1=PDF, 2=BB)
        2,          //  1: PS version
        0,          //  2: PS Size: 0=1:1, 1=res chunk, 2=media size
        255,        //  3: PS alpha mix red
        255,        //  4: PS alpha mix green
        255,        //  5: PS alpha mix blue
        2,          //  6: PDF Index of version
        0,          //  7: PDF Index of fit entry
        0,          //  8: PDF size
        255,        //  9: PDF alpha mix red
        255,        // 10: PDF alpha mix green
        255,        // 11: PDF alpha mix blue
        0,          // 12: PS options tab
        0,          // 13: PDF options tab
        0,          // 14: PDF predictor
    };

    /** Flags which file types to handle. */
    protected boolean[] fileTypesToHandle = {
        true,   //  0: PNG
        false,  //  1: JPEG
        false,  //  2: NetPBM
        false,  //  3: TIFF
    };

    public BmepsController(DkTool d, DkGuiTool g) {
        dkt = d; dkgt = g;
        dkt.getMultiplePreferenceStrings(pkStr, myStr);
        dkt.getMultiplePreferenceInts(pkInt, myInt);
        dkt.getMultiplePreferenceBools(pkBool, myBool);
        int state = 0;
        int i;
        ArrayList<String> alb = new ArrayList<String>();
        ArrayList<String> alc = new ArrayList<String>();
        try {
            ProcessBuilder pbuild = new ProcessBuilder(cmdToReadConfigFile);
            Process p = pbuild.start();
            InputStreamReader isr = null;
            isr = new InputStreamReader(p.getInputStream());
            BufferedReader in = null;
            in = new BufferedReader(isr);
            for(String s; (s = in.readLine()) != null; ) {
                s = s.trim();
                if(s.startsWith("# @")) {
                    if(s.contains("@BUILT-IN")) {
                        state = 0;
                    } else {
                        if(s.contains("@CONFIG-FILE")) {
                            state = 1;
                        } else {
                            if(s.startsWith("# @FILE-TYPES ")) {
                                fileTypesToHandle[0] = false;
                                fileTypesToHandle[1] = false;
                                fileTypesToHandle[2] = false;
                                fileTypesToHandle[3] = false;
                                if(s.contains("PNG")) {
                                    fileTypesToHandle[0] = true;
                                }
                                if(s.contains("JPEG")) {
                                    fileTypesToHandle[1] = true;
                                }
                                if(s.contains("NetPBM")) {
                                    fileTypesToHandle[2] = true;
                                }
                                if(s.contains("TIFF")) {
                                    fileTypesToHandle[3] = true;
                                }
                            }
                        }
                    }
                } else {
                    if(s.startsWith("[")) {
                        s = s.substring(1);
                        i = s.indexOf(']');
                        if(i >= 0) {
                            s = s.substring(0, i);
                            i = s.indexOf(':');
                            if(i >= 0) {
                                s = s.substring(0, i);
                            }
                            switch(state) {
                                case 1: {
                                    alc.add(s);
                                } break;
                                default: {
                                    alb.add(s);
                                } break;
                            }
                        }
                    }
                }
            }
            if(in != null) try { in.close(); } catch(Exception e) {}
            in = null;
            if(isr != null) try { isr.close(); } catch(Exception e) {}
            isr = null;
            isr = new InputStreamReader(p.getErrorStream());
            in = new BufferedReader(isr);
            for(String s; (s = in.readLine()) != null; ) {}
            if(in != null) try { in.close(); } catch(Exception e) {}
            in = null;
            if(isr != null) try { isr.close(); } catch(Exception e) {}
            isr = null;
            p.waitFor();
            if(p.exitValue() == 0) {
                bmepsFound = true;
            }
        } catch(Exception e) { bmepsFound = false; }
        Object[] oa = alb.toArray();
        builtinConfigurations = new String[oa.length];
        for(i = 0; i < oa.length; i++) {
            builtinConfigurations[i] = new String((String)(oa[i]));
        }
        oa = alc.toArray();
        fileConfigurations = new String[oa.length];
        for(i = 0; i < oa.length; i++) {
            fileConfigurations[i] = new String((String)(oa[i]));
        }
    }


    public void saveSettings() {
        dkt.setMultiplePreferenceBools(pkBool, myBool);
        dkt.setMultiplePreferenceInts(pkInt, myInt);
        dkt.setMultiplePreferenceStrings(pkStr, myStr);
        dkt.flushPreferences();
    }



    public void setString(int ind, String s) {
        if((ind >= 0) && (ind < myStr.length)) {
            synchronized(this) {
                myStr[ind] = s;
            }
        }
    }

    public String getString(int ind) {
        String back = null;
        if((ind >= 0) && (ind < myStr.length)) {
            synchronized(this) {
                back = myStr[ind];
            }
        }
        return back;
    }

    public String[] getBuiltinConfigurations() {
        return builtinConfigurations;
    }

    public String[] getFileConfigurations() {
        return fileConfigurations;
    }

    public boolean getData(boolean[] b, int[] iv, String[] s) {
        boolean back = false;
        int i;
        if(b.length >= myBool.length) {
            if(iv.length >= myInt.length) {
                if(s.length >= myStr.length) {
                    synchronized(this) {
                        for(i = 0; i < myBool.length; i++) {
                            b[i] = myBool[i];
                        }
                        for(i = 0; i < myInt.length; i++) {
                            iv[i] = myInt[i];
                        }
                        for(i = 0; i < myStr.length; i++) {
                            if(myStr[i] != null) { s[i] = new String(myStr[i]); }
                            else { s[i] = null; }
                        }
                    }
                    back = true;
                }
            }
        }
        return back;
    }


    public boolean putData(boolean[] b, int[] iv, String[] s) {
        boolean back = false;
        int i;
        if(b.length >= myBool.length) {
            if(iv.length >= myInt.length) {
                if(s.length >= myStr.length) {
                    synchronized(this) {
                        for(i = 0; i < myBool.length; i++) {
                            myBool[i] = b[i];
                        }
                        for(i = 0; i < myInt.length; i++) {
                            myInt[i] = iv[i];
                        }
                        for(i = 0; i < myStr.length; i++) {
                            if(s[i] != null) { myStr[i] = new String(s[i]); }
                            else { myStr[i] = null; }
                        }
                    }
                    back = true;
                }
            }
        }
        return back;
    }
    
    
    
    private String findOutputConfiguration(boolean[] b, int[] i, String[] s) {
        String back = null;
        if(b[2]) {  // full-custom
            try {
                back = baseDrivers[ i[0] ];
            } catch(Exception e) { back = baseDrivers[0]; }
        } else {    // predefined
            if(b[3]) {  // config-file
                back = s[4];
            } else {    // built-in
                back = s[1];
            }
        }
        return back;
    }



    private String findOutputSuffix(boolean[] b, int[] i, String[] s) {
        String back = null;
        String oc = findOutputConfiguration(b, i, s);
        if(oc != null) {
            if(oc.contains("pdf")) {
                back = outputSuffixes[1];
            } else {
                if(oc.contains("eps")) {
                    back = outputSuffixes[0];
                } else {
                    if(oc.contains("bb")) {
                        back = outputSuffixes[2];
                    } else {
                        back = outputSuffixes[3];
                        if(b[4]) {
                            back = outputSuffixes[0];
                        }
                    }
                }
            }
        } else {
            back = outputSuffixes[0];
        }
        return back;
    }


    private
    String
    createOutputFilename(String ifn, boolean[] b, int[] i, String[] s)
    {
        String back = null;
        back = dkt.replaceSuffix(ifn, findOutputSuffix(b, i, s));
        return back;
    }


    private
    void createListOfFiles(ArrayList<String> a, boolean[] b, int[] i, String[] s)
    {
        boolean caseSensitive = dkt.getPfi().getFilenamesCaseSensitive();
        boolean useBackslash = dkt.getPfi().getFilenamesSeparatorBackslash();
        boolean can_handle;
        boolean must_handle;
        String suf;
        String ifn;
        String ofn;
        String dirname = s[0];
        if(dirname == null) { dirname = "."; }
        File currentDirectory = new File(dirname);
        if(currentDirectory.exists()) {
            if(currentDirectory.isDirectory()) {
                String[] allFileNames = currentDirectory.list();
                for(String filename: allFileNames) {
                    can_handle = false; must_handle = false;
                    suf = dkt.getFilenameSuffix(filename);
                    if(useBackslash) {
                        ifn = dirname + "\\" + filename;
                    } else {
                        ifn = dirname + "/" + filename;
                    }
                    File fifn = new File(ifn);
                    if(fifn.exists() && fifn.isFile()) {
                        if(suf != null) {
                            if((!can_handle) && fileTypesToHandle[0]) {
                                if(suf.equalsIgnoreCase(".png")) {
                                    can_handle = true;
                                }
                            }
                            if((!can_handle) && fileTypesToHandle[1]) {
                                if(suf.equalsIgnoreCase(".jpg")) {
                                    can_handle = true;
                                }
                                if(suf.equalsIgnoreCase(".jpeg")) {
                                    can_handle = true;
                                }
                            }
                            if((!can_handle) && fileTypesToHandle[2]) {
                                if(suf.equalsIgnoreCase(".ppm")) {
                                    can_handle = true;
                                }
                                if(suf.equalsIgnoreCase(".pbm")) {
                                    can_handle = true;
                                }
                                if(suf.equalsIgnoreCase(".pgm")) {
                                    can_handle = true;
                                }
                                if(suf.equalsIgnoreCase(".pnm")) {
                                    can_handle = true;
                                }
                            }
                            if((!can_handle) && fileTypesToHandle[3]) {
                                if(suf.equalsIgnoreCase(".tif")) {
                                    can_handle = true;
                                }
                                if(suf.equalsIgnoreCase(".tiff")) {
                                    can_handle = true;
                                }
                            }
                        }
                        if(can_handle) {
                            must_handle = true;
                            ofn = createOutputFilename(ifn, b, i, s);
                            if(b[0]) {
                                File fofn = new File(ofn);
                                if(fofn.exists()) {
                                    if(fofn.isFile()) {
                                        if(fofn.lastModified() > fifn.lastModified())
                                        {
                                            must_handle = false;
                                        }
                                    }
                                }
                            }
                            if(must_handle) {
                                a.add(ifn);
                            }
                        }
                    }
                }
            }
        }
    }


    private String colorDoubleString(int i) {
      int j = i;
      if(i < 0) j = 0;
      if(i > 255) j = 255;
      double d = (double)j / 255.0;
      String back = (new Double(d)).toString();
      return back;
    }


    public TreeMap<String,String[]> createListOfCommands() {
        TreeMap<String,String[]> back = new TreeMap<String,String[]>();
        ArrayList<String> al = new ArrayList<String>();
        ArrayList<String> nc = null;
        String ifn;
        String ofn;
        String sfn;
        boolean[] b = new boolean[myBool.length];
        int[]     i = new int[myInt.length];
        String[]  s = new String[myStr.length];
        int j;
        synchronized(this) {
            for(j = 0; j < myBool.length; j++) { b[j] = myBool[j]; }
            for(j = 0; j < myInt.length; j++) { i[j] = myInt[j]; }
            for(j = 0; j < myStr.length; j++) {
                if(myStr[j] != null) { s[j] = new String(myStr[j]); }
                else { s[j] = null; }
            }
        }
        createListOfFiles(al, b, i, s);
        for(Object o: al.toArray()) {
            ifn = (String)o;
            ofn = createOutputFilename(ifn, b, i, s);
            File fifn = new File(ifn);
            sfn = fifn.getName();
            nc = new ArrayList<String>();
            nc.add(bmepsOptions[0]);
	    nc.add(bmepsOptions[1]);
            if(b[2]) {  // full-custom
	        StringBuilder sb = new StringBuilder();
                switch(i[0]) {
                    case 2: {   // BB
		      sb.append(baseDrivers[2]);
                    } break;
                    case 1: {   // PDF
		      sb.append(baseDrivers[1]);
		      if(b[1]) {
		        sb.append(bmepsOptions[3]);
		      } else {
		        sb.append(bmepsOptions[4]);
			sb.append(pdfVersions[i[6]]);
			if(i[7] > 0) {
			  sb.append(bmepsOptions[42]);
			  sb.append(pdfFit[i[7] - 1]);
			}
			sb.append(bmepsOptions[b[38] ? 9 : 10]);
			if(b[38]) {
			  sb.append(bmepsOptions[b[39] ? 11 : 12]);
			}
			sb.append(bmepsOptions[b[40] ? 21 : 22]);
			if(!(b[40])) {
			  boolean is_first = true;
			  for(int k = 0; k < 16; k++) {
			    if(b[41 + k]) {
			      sb.append(bmepsOptions[is_first ? 23 : 14]);
			      is_first = false;
			      sb.append((new Integer(k)).toString());
			    }
			  }
			}
			switch(i[8]) {
			  case 2: {
			    if(s[3] != null) {
			      sb.append(bmepsOptions[25]);
			      sb.append(s[3]);
			    }
			  } break;
			  case 1: {
			    sb.append(bmepsOptions[24]);
			  } break;
			}
			if(i[6] > 1) {
			  sb.append(bmepsOptions[b[57] ? 43 : 44]);
			  if(b[57]) {
			    sb.append(bmepsOptions[b[58] ? 45 : 46]);
			  }
			}
			if(!((i[6] > 1) && (b[57]))) {
			  sb.append(bmepsOptions[b[59] ? 26 : 27]);
			  if(b[59]) {
			    sb.append(bmepsOptions[28]);
			    sb.append(colorDoubleString(i[9]));
			    sb.append(bmepsOptions[14]);
			    sb.append(colorDoubleString(i[10]));
			    sb.append(bmepsOptions[14]);
			    sb.append(colorDoubleString(i[11]));
			    sb.append(bmepsOptions[b[60] ? 29 : 30]);
			  }
			  sb.append(bmepsOptions[b[61] ? 31 : 32]);
			  if(b[61]) {
			    sb.append(bmepsOptions[b[62] ? 34 : 33]);
			  }
			}
		      }
                      if(myInt[14] > 0) {
                          if(myInt[14] < 6) {
                              sb.append(bmepsOptions[57]);
                              sb.append(bmepsOptions[58 + myInt[14]]);
                          }
                      }
                    } break;
                    default: {  // PS
                      sb.append(baseDrivers[b[4] ? 3 : 0]);
		      if(b[1]) {
		        sb.append(bmepsOptions[3]);
		      } else {
		        sb.append(bmepsOptions[4]);
			sb.append((new Integer(i[1])).toString());
			sb.append(bmepsOptions[b[5] ? 5 : 6]);
			if(i[1] > 1) {
			  sb.append(bmepsOptions[b[6] ? 7 : 8]);
			  sb.append(bmepsOptions[b[7] ? 9 : 10]);
			  if(b[7]) {
			    sb.append(bmepsOptions[b[8] ? 11 : 12]);
			  }
			  boolean is_first = true;
			  if(b[9]) {
			    sb.append(bmepsOptions[is_first ? 13 : 14]);
			    sb.append(bmepsOptions[15]);
			    is_first = false;
			  }
			  if(b[10]) {
			    sb.append(bmepsOptions[is_first ? 13: 14]);
			    is_first = false;
			    sb.append(bmepsOptions[16]);
			  }
			  if(i[1] > 2) {
			    if(b[12]) {
			      sb.append(bmepsOptions[is_first ? 13: 14]);
			      is_first = false;
			      sb.append(bmepsOptions[17]);
			    }
			  }
			  if(b[13]) {
			    sb.append(bmepsOptions[is_first ? 13: 14]);
			    sb.append(bmepsOptions[20]);
			  }
			  if(b[10]) {
			    sb.append(bmepsOptions[b[11] ? 18 : 19]);
			  }
			  if(b[13]) {
			    sb.append(bmepsOptions[b[14] ? 21 : 22]);
			    if(!(b[14])) {
			      is_first = true;
			      for(int k = 0; k < 16; k++) {
			        if(b[15 + k]) {
				  sb.append(bmepsOptions[is_first ? 23 : 14]);
				  is_first = false;
				  sb.append((new Integer(k)).toString());
				}
			      }
			    }
			  }
			  switch(i[2]) {
			    case 2: {
			      if(myStr[2] != null) {
			        sb.append(bmepsOptions[25]);
				sb.append(myStr[2]);
			      }
			    } break;
			    case 1: {
			      sb.append(bmepsOptions[25]);
			    } break;
			  }
			  sb.append(bmepsOptions[b[31] ? 26 : 27]);
			  if(b[31]) {
			    sb.append(bmepsOptions[28]);
			    sb.append(colorDoubleString(i[3]));
			    sb.append(bmepsOptions[14]);
			    sb.append(colorDoubleString(i[4]));
			    sb.append(bmepsOptions[14]);
			    sb.append(colorDoubleString(i[5]));
			    sb.append(bmepsOptions[b[32] ? 29 : 30]);
			  }
			  sb.append(bmepsOptions[b[33] ? 31 : 32]);
			  if(b[33]) {
			    sb.append(bmepsOptions[b[34] ? 34 : 33]);
			  }
			  sb.append(bmepsOptions[b[35] ? 35 : 36]);
			  if(b[35]) {
			    sb.append(bmepsOptions[b[36] ? 37 : 38]);
			  }
			  sb.append(bmepsOptions[39]);
			  sb.append(bmepsOptions[b[37] ? 40 : 41]);
			}
		      }
                    } break;
                }
		nc.add(bmepsOptions[2]);
		nc.add(new String(sb));
            } else {    // predefined
                if(b[3]) {  // config-file
                    if(s[4] != null) {
                        nc.add(bmepsOptions[2]);
                        nc.add(s[4]);
                    }
                } else {    // built-in
                    if(s[1] != null) {
                        nc.add(bmepsOptions[2]);
                        nc.add(s[1]) ;
                    }
                }
            }
            nc.add(ifn);
            nc.add(ofn);
	    String lang = dkt.getLanguage();
	    if(lang.indexOf('.') < 0) {
	      lang = lang + bmepsOptions[49];
	    }
	    nc.add(bmepsOptions[47]);
	    nc.add(bmepsOptions[48] + lang);
	    nc.add(bmepsOptions[50]);
	    nc.add(bmepsOptions[51]);
	    nc.add(bmepsOptions[52]);
	    nc.add(bmepsOptions[53]);
	    nc.add(bmepsOptions[54]);
	    nc.add(bmepsOptions[55]);
	    nc.add(bmepsOptions[56]);
            Object[] oa = nc.toArray();
            String[] cma = new String[oa.length];
            for(int ind = 0; ind < oa.length; ind++) {
                cma[ind] = new String((String)(oa[ind]));
            }
            back.put(sfn, cma);
        }
        return back;
    }

}
