/*
Copyright (c) 2009, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.tools;

import java.awt.Rectangle;
import java.awt.Dimension;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.Toolkit;

import java.net.URI;
import java.net.URL;

import javax.swing.JFrame;
import javax.swing.JOptionPane;

/**
 * The DkFrame class saves the window position of the last window
 * closed.
 * After starting the application again the choosePosition() function
 * can be used after pack() to attempt to restore the last position.
 *
 * The cleanup() method is invoked if the window is closed.
 * The parameter isLast of this method indicates whether or not a window
 * is the ``last of it's kind''. May be you want to save program settings
 * if the last JFrame of the application is closed.
 *
 * @author krause
 */
abstract public class DkFrame
    extends JFrame
    implements WindowListener
{
    /** Preference keys to save/restore window size and position. */
    private static String[] pkInt = {
        /*   0 */   "ui.window.x",
        /*   1 */   "ui.window.y",
        /*   2 */   "ui.window.w",
        /*   3 */   "ui.window.h",
    };

    /** Screen width in pixels. */
    protected int screenWidth = 0;

    /** Screen height in pixels. */
    protected int screenHeight = 0;

    /** Preferences manager. */
    protected DkTool dkt = null;

    /** GUI tool. */
    protected DkGuiTool dkgt = null;

    /** X and Y position, width and height. */
    private int[] xy = { -1, -1, -1, -1 };

    /** Number of windows of this class currently open. */
    private static int numberOfInstances = 0;

    /** Space between components. */
    protected int border = 5;

    /** GUI dialog. */
    protected DkGuiDialog guiDialog = null;

    /** Dialog box for messages. */
    protected MessageDialog messageDialog = null;

    /**
     * Increase number of open instances.
     * @return  New number of open instances.
     */
    private synchronized int addOneInstance() {
        numberOfInstances++;
        return numberOfInstances;
    }


    /**
     * Decrease number of open instances.
     * @return New number of open instances.
     */
    private synchronized int removeOneInstance() {
        numberOfInstances--;
        if(numberOfInstances < 0) { numberOfInstances = 0; }
        return numberOfInstances;
    }

    /**
     * Constructor.
     * @param a Application name.
     * @param d Tool object to manage preferences.
     */
    public DkFrame(String a, DkTool d, DkGuiTool g) {
        super(a);
	dkt = d; dkgt = g;
        addOneInstance();
        addWindowListener(this);
        dkt.getMultiplePreferenceInts(pkInt, xy);
        border = dkgt.getComponentBorder();
        Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
        screenWidth = dim.width; screenHeight = dim.height;
    }


    /**
     * Choose window position either by retrieving the position
     * of the last window closed or by placing the window centered
     * on the screen.
     * Must be called after pack().
     */
    public void choosePosition() {
        Rectangle rv = new Rectangle();
        getBounds(rv);
        if(xy[2] >= rv.width) {
            if(xy[3] >= rv.height) {
                this.setSize(xy[2], xy[3]);
                getBounds(rv);
            }
        }
        if(xy[0] < 0) {
            xy[0] = (screenWidth - rv.width) / 2;
        }
        if(xy[1] < 0) {
            xy[1] = (screenHeight - rv.height) / 2;
        }
        if((xy[0] + rv.width) > screenWidth) {
            xy[0] = screenWidth - rv.width;
        }
        if((xy[1] + rv.height) > screenHeight) {
            xy[1] = screenHeight - rv.height;
        }
        if(xy[0] < 0) { xy[0] = 0; }
        if(xy[1] < 0) { xy[1] = 0; }
        setLocation(xy[0], xy[1]);
    }

    /** Get DkTool object. */
    protected DkTool getTool() {
        return dkt;
    }

    /** Get DkGuiTool object. */
    protected DkGuiTool getGuiTool() {
        return dkgt;
    }

    /**
     * Show GUI setup dialog.
     */
    protected void showGuiDialog() {
        DkGuiDialog d;
	synchronized(this) {
	    d = guiDialog;
	}
	if(d == null) {
	    d = new DkGuiDialog(this, dkt, dkgt);
	    dkt.chooseModalPosition(screenWidth, screenHeight, this, d);
	    synchronized(this) {
	        guiDialog = d;
	    }
	}
	d.dataGet();
	d.setModal(true);
	d.setVisible(true);
    }

    /**
     * Show message dialog.
     * @param t Title for dialog.
     * @param m Message text.
     * @param html  Flag: Text is HTML.
     */
    protected void showMessageDialog(String t, String m, boolean html) {
        MessageDialog d;
	boolean must_set = false;
	synchronized(this) {
	    d = messageDialog;
	}
	if(d == null) {
	    d = new MessageDialog(this, dkt, dkgt);
            synchronized(this) {
                messageDialog = d;
            }
	    must_set = true;
	}
        if(must_set) {
	    dkt.chooseModalPosition(screenWidth, screenHeight, this, d);
	}
	d.setTitle(t);
        d.setHtml(html);
	d.setEditorText(m);
	d.setModal(true);
	d.setVisible(true);
        d.repaint();
    }


    /**
     * Show message dialog.
     * @param t     Title for dialog
     * @param u     URL of page to show.
     * @param html  Flag: Text is HTML.
     * @return      Flag to indicate success.
     */
    protected boolean showMessageDialogForURL(String t, URL u, boolean html) {
        boolean back = false;
        MessageDialog d;
        boolean must_set = false;
        synchronized(this) {
            d = messageDialog;
        }
        if(d == null) {
            d = new MessageDialog(this, dkt, dkgt);
            synchronized(this) {
                messageDialog = d;
            }
            must_set = true;
        }
        if(must_set) {
            dkt.chooseModalPosition(screenWidth, screenHeight, this, d);
        }
        d.setTitle(t);
        d.setHtml(html);
        back = d.setEditorPage(u);
        d.setModal(true);
        d.setVisible(true);
        d.repaint();
        return back;
    }



    /**
     * Show message in dialog box or message dialog.
     * @param	t	Dialog box title.
     * @param	m	Message.
     * @param	l	Message level (JOptionPane xxx_MESSAGE).
     * @param	f	Flag: Message is HTML formatted.
     */
    public void internalMessageBox(String t, String m, int l, boolean f)
    {
        boolean failed = false;
	try {
	    JOptionPane.showMessageDialog(this, m, t, l);
	} catch(Exception e) { failed = true; }
	if(failed) {
	    showMessageDialog(t, m, f);
	}
    }



    /**
     * Show information message dialog box.
     * @param	t	Dialog box title.
     * @param	m	Message.
     * @param	f_html	Flag: Message is HTML formatted.
     */
    public void msgInfo(String t, String m, boolean f_html) {
        internalMessageBox(
	    t, m, JOptionPane.INFORMATION_MESSAGE, f_html
	);
    }

    
    
    /**
     * Show warning message dialog box.
     * @param	t	Dialog box title.
     * @param	m	Message.
     * @param	f_html	Flag: Message is HTML formatted.
     */
    public void msgWarning(String t, String m, boolean f_html) {
        internalMessageBox(
	    t, m, JOptionPane.WARNING_MESSAGE, f_html
	);
    }



    /**
     * Show error message dialog box.
     * @param	t	Dialog box title.
     * @param	m	Message.
     * @param	f_html	Flag: Message is HTML formatted.
     */
    public void msgError(String t, String m, boolean f_html) {
        internalMessageBox(
	    t, m, JOptionPane.ERROR_MESSAGE, f_html
	);
    }



    /**
     * Cleanup method, invoked when the window is closed.
     *
     * @param isFinalWindow
     * Indicates whether or not this was the last visible
     * window.
     */
    abstract public void cleanup(boolean isFinalWindow);



    /**
     * Callback method, invoked when a window is closed.
     *
     * @param e
     * The window event indicating that the window is closed.
     */
    @Override
    public void windowClosing(WindowEvent e) {
        int now = removeOneInstance();
        removeWindowListener(this);
        setVisible(false);
        if(now == 0) {
            Rectangle rv = new Rectangle();
            getBounds(rv);
            xy[0] = rv.x;
            xy[1] = rv.y;
            xy[2] = rv.width;
            xy[3] = rv.height;
            rv = null;
            dkt.setMultiplePreferenceInts(pkInt, xy);
            dkt.flushPreferences();
        }
        cleanup((now == 0) ? true : false);
        removeAll();
        dkt = null;
        dkgt = null;
        dispose();
        if(now == 0) {
            System.gc();
            System.runFinalization();
            try { System.exit(0); } catch(Exception exc) {}
        }
    }

    /** Unused window event handler. */
    @Override
    public void windowActivated(WindowEvent e) {}

    /** Unused window event handler. */
    @Override
    public void windowClosed(WindowEvent e) {}

    /** Unused window event handler. */
    @Override
    public void windowDeactivated(WindowEvent e) {}

    /** Unused window event handler. */
    @Override
    public void windowDeiconified(WindowEvent e) {}

    /** Unused window event handler. */
    @Override
    public void windowIconified(WindowEvent e) {}

    /** Unused window event handler. */
    @Override
    public void windowOpened(WindowEvent e) {}

}

