/*
Copyright (c) 2009, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/


package dirk_krause.tools;

import java.util.Enumeration;
import java.util.HashSet;
import java.awt.Font;
import java.awt.Toolkit;
import javax.swing.UIManager;
import javax.swing.UIDefaults;
import javax.swing.LookAndFeel;
import javax.swing.plaf.FontUIResource;

/**
 * Dirk Krause's general tool class.
 * @author krause
 */
public class DkGuiTool {
    /** Used to manage preferences. */
    private DkTool dkt = null;

    /** Flag: First instance, must initialize. */
    private static boolean is_first = true;

    /** Static strings used by the class. */
    private static final String[] stt = {
        /*   0 */   ", "  ,
    };

    /** Settings (integer numbers). */
    private int[] myInt = {
        5,  //   0: Component border
        0,  //   1: Minimum font size
        0,  //   2: look-and-feel index
    };

    /** Preference keys to save/restore the integer numbers above. */
    private String[] myPkInt = {
        "ui.comonent-border",
        "ui.font-size.minimum",
        "ui.look-and-feel",
    };

    /** Settings (texts). */
    private String[] myStr = {
        null,   //   0: Font replacement, old font
        null,   //   1: Font replacement, new font
    };

    /** Preference keys to save/restore the texts above. */
    private String[] myPkStr = {
        "ui.font-replacement.old",
        "ui.font-replacement.new",
    };

    /** Settings (boolean values). */
    private boolean[] myBool = {
        false,  //   0: Replace fonts
    };

    /** Preference keys to save/restore the values above. */
    private String[] myPkBool = {
        "ui.font-replacement.enabled",
    };

    /** Fonts used before font replacements. */
    private String fontsBeforeReplacement = "";

    /** Fonts used after font replacements. */
    private String fontsAfterReplacement = "";

    /**
     * Constructor.
     * @param d             DkTool object.
     * @param useToolkit    Automatic font size calculation allowed.
     */
    public DkGuiTool(DkTool d, boolean useToolkit) {
        dkt = d;
        if(useToolkit) {
            myInt[1] =
            (14 * Toolkit.getDefaultToolkit().getScreenResolution()) / 98;
        }
        d.getMultiplePreferenceStrings(myPkStr, myStr);
        d.getMultiplePreferenceInts(myPkInt, myInt);
        d.getMultiplePreferenceBools(myPkBool, myBool);
	initialize();
    }

    /**
     * Constructor (simple call).
     * @param d DkTool object.
     */
    public DkGuiTool(DkTool d) {
        this(d, true);
    }

    /**
     * Modify the fonts (replace font families and enforce minimum size).
     * @param d     User interfacde defaults set.
     * @param s1    Collection of fonts before replacement.
     * @param s2    Collection of fonts after replacement.
     */
    private void fontModifications(
        UIDefaults d, HashSet<String> s1, HashSet<String> s2
    )
    {
        Enumeration en = d.keys();			// defaults keys
        HashSet<Object> hs = new HashSet<Object>();	// objects we process
        boolean must_replace;			// flag: must replace
        boolean do_font_replacements = false;	// must do font replacements
        Object	k = null;			// key object
        Object	v = null;			// value object
        Font	f = null;			// font
        FontUIResource	fuir = null;		// FontUIResource
        String	font_name = null;		// font name
        int		font_style = 0;			// font style
        int		font_size = 0;			// font size
        String	old_font_name = null;		// old font name
        String	new_font_name = null;		// new font name
        int		min_font_size = 0;
        do_font_replacements = getReplaceFonts();
        if(do_font_replacements) {
            old_font_name = getOldFont();
            new_font_name = getNewFont();
        }
        min_font_size = getMinimumFontSize();
        while(en.hasMoreElements()) {
            try {
                k = en.nextElement();
                v = d.get(k);
                if(v != null) {
                    if((v instanceof Font) || (v instanceof FontUIResource)) {
                        hs.add(k);
                    }
                }
            } catch(Exception e) {}
            Object[] keyArray;
            keyArray = hs.toArray();
            for(int i = 0; i < keyArray.length; i++) {
                must_replace = false;
                k = keyArray[i];
                try {
                    v = d.get(k);
                    if(v != null) {
                        if(v instanceof Font) {
                            f = (Font)v;
                            font_name = f.getName();
                            font_style = f.getStyle();
                            font_size = f.getSize();
                            s1.add(
                                new String(
                                    font_name + " " +
                                    (new Integer(font_size)).toString()
                                )
                            );
                            if(do_font_replacements) {
                                if(font_name.equals(old_font_name)) {
                                    font_name = new_font_name;
                                    must_replace = true;
                                }
                            }
                            if(min_font_size > 0) {
                                if(font_size < min_font_size) {
                                    font_size = min_font_size;
                                    must_replace = true;
                                }
                            }
                            if(must_replace) {
                                d.put(
                                    k,
                                    (Object)(new Font(
                                        font_name, font_style, font_size)
                                    )
                                );
                            }
                            s2.add(
                                new String(
                                    font_name + " " +
                                    (new Integer(font_size)).toString()
                                )
                            );
                        } else {
                            if(v instanceof FontUIResource) {
                                fuir = (FontUIResource)v;
                                font_name = fuir.getName();
                                font_style = fuir.getStyle();
                                font_size = fuir.getSize();
                                s1.add(
                                    new String(
                                        font_name + " " +
                                        (new Integer(font_size)).toString()
                                    )
                                );
                                if(do_font_replacements) {
                                    if(font_name.equals(old_font_name)) {
                                        font_name = new_font_name;
                                        must_replace = true;
                                    }
                                }
                                if(min_font_size > 0) {
                                    if(font_size < min_font_size) {
                                        font_size = min_font_size;
                                        must_replace = true;
                                    }
                                }
                                if(must_replace) {
                                    d.put(
                                        k,
                                        (Object)(new FontUIResource(
                                            font_name, font_style, font_size
                                        ))
                                    );
                                }
                                s2.add(
                                    new String(
                                        font_name + " " +
                                        (new Integer(font_size)).toString()
                                    )
                                );
                            }
                        }
                    }
                } catch(Exception e) {}
            }
        }
    }

    /**
     * Convert a font name hashset to a string.
     * @param shs   The hash set containing all font/size texts.
     * @return      A string which can be shown in a JLabel.
     */
    private String fontHashSetToString(HashSet<String> shs) {
        StringBuilder sb = new StringBuilder();
        Object k = null;
        boolean is_not_first = false;
        Object keyArray[] = shs.toArray();
        for(int i = 0; i < keyArray.length; i++) {
            k = keyArray[i];
            if(is_not_first) {
                sb.append(stt[0]);
            }
            sb.append((String)k);
            is_not_first = true;
        }
        String back = sb.toString();
        return back;
    }

    /** Set the string for the font names before replacement.
     *  @param s    New string value.
     */
    private void setFontsUsed1(String s) {
        synchronized(this) {
            fontsBeforeReplacement = s;
        }
    }

    /**
     * Setthe string for the font names after replacement.
     * @param s New string value.
     */
    private void setFontsUsed2(String s) {
        synchronized(this) {
            fontsAfterReplacement = s;
        }
    }

    /**
     * Get string with font names before replacement.
     * @return  String listing old font names.
     */
    public String getFontsBeforeReplacement() {
        String back;
        synchronized(this) {
            back = fontsBeforeReplacement;
        }
        if(back != null) { back = new String(back); }
        return back;
    }

    /**
     * Get string with font names after replacement.
     * @return  String listing the new font names.
     */
    public String getFontsAfterReplacement() {
        String back;
        synchronized(this) {
            back = fontsAfterReplacement;
        }
        if(back != null) { back = new String(back); }
        return back;
    }

    /**
     * Run initialization.
     */
    private void doInitialize() {
        HashSet<String> s1 = new HashSet<String>();	// old fonts
        HashSet<String> s2 = new HashSet<String>();	// new fonts
        switch(getLafIndex()) {
            case 0: {   // System look and feel
                String s = UIManager.getSystemLookAndFeelClassName();
                try {
                    UIManager.setLookAndFeel(s);
                } catch(Exception e) {}
                LookAndFeel laf = UIManager.getLookAndFeel();
                if(laf != null) {
                    UIDefaults defs = laf.getDefaults();
                    fontModifications(defs, s1, s2);
                }
            } break;
            case 1: {   // Java LAF

            } break;
        }
        // step 2: correct defaults
        UIDefaults d2 = UIManager.getDefaults();
        fontModifications(d2, s1, s2);
        // step 3: set font strings
        setFontsUsed1(fontHashSetToString(s1));
        setFontsUsed2(fontHashSetToString(s2));
    }

    /**
     * Check whether we still need to initialize or not.
     * @return  Flag: Must initialize.
     */
    private synchronized boolean mustInitialize() {
        boolean back;
	back = is_first; is_first = false;
	return back;
    }


    /**
     * Attempt to initialize GUI and toolkit.
     */
    public void initialize() {
        if(mustInitialize()) {
	    doInitialize();
	}
    }

    /**
     * Save GUI settings.
     * @param stage Destination (0=application, 1=app+group, 2=app+group+def).
     * @param s     Flag: Attempt settings in system pref nodes too.
     */
    public void save(int stage, boolean s) {
        boolean g = false;
        boolean c = false;
        if(stage > 0) { g = true; }
        if(stage > 1) { c = true; }
        synchronized(this) {
            if(dkt != null) {
                dkt.setMultiplePreferenceBools(myPkBool, myBool, g, c, s);
                dkt.setMultiplePreferenceInts(myPkInt, myInt, g, c, s);
                dkt.setMultiplePreferenceStrings(myPkStr, myStr, g, c, s);
            }
        }
        dkt.flushPreferences();
    }

    /**
     * Set component border.
     * @param i New number of pixels for space between components.
     */
    public void setComponentBorder(int i) {
        synchronized(this) {
	    myInt[0] = i;
	}
    }

    /**
     * Get border setting.
     * @return  Number of pixels for space between components.
     */
    public int getComponentBorder() {
        int back = 0;
        synchronized(this) {
            back = myInt[0];
        }
        return back;
    }

    /**
     * Set the minimum font size to enforce.
     * @param i New minimum font size.
     */
    public void setMinimumFontSize(int i) {
    	synchronized(this) {
	    myInt[1] = i;
	}
    }

    /**
     * Get minimum font size to enforce.
     * @return  The minimum font size.
     */
    public int getMinimumFontSize() {
    	int back = 0;
	synchronized(this) {
	    back = myInt[1];
	}
	return back;
    }

    /**
     * Set index of look-and-feel.
     * @param i New index value.
     */
    public void setLafIndex(int i) {
        int li;
	li = i;
	if(li > 2) {
	    li = 2;
	}
	synchronized(this) {
	    myInt[2] = li;
	}
    }

    /**
     * Get look-and-feel index.
     * @return  Current LAF index.
     */
    public int getLafIndex() {
        int back = 0;
	synchronized(this) {
	    back = myInt[2];
	}
	return back;
    }

    /**
     * Set font family name to be replaced before creating the GUI.
     * @param o Font family to replace.
     */
    public void setOldFont(String o) {
        String s = new String(o);
	synchronized(this) {
	    myStr[0] = s;
	}
    }

    /**
     * Get font family name to replace before creating the GUI.
     * @return  The font family name.
     */
    public String getOldFont() {
        String back;
	synchronized(this) {
	    back = myStr[0];
	}
	if(back != null) { back = new String(back); }
	return back;
    }

    /**
     * Set replacement font family name.
     * @param n New replacement font family name.
     */
    public void setNewFont(String n) {
        String s = new String(n);
	synchronized(this) {
	    myStr[1] = s;
	}
    }

    /**
     * Get the replacement font family name.
     * @return  The font family name.
     */
    public String getNewFont() {
        String back;
	synchronized(this) {
	    back = myStr[1];
	}
	if(back != null) { back = new String(back); }
	return back;
    }

    /**
     * Enable/disable font family replacement before creating the GUI.
     * @param b Flag to enable/disable.
     */
    public void setReplaceFonts(boolean b) {
        synchronized(this) {
	    myBool[0] = b;
	}
    }

    /**
     * Get font replacement setting.
     * @return  Flag whether or not to do font family replacements.
     */
    public boolean getReplaceFonts() {
        boolean back;
	synchronized(this) {
	    back = myBool[0];
	}
	return back;
    }

}
