/*
Copyright (c) 2009, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.tools;

import java.util.prefs.Preferences;
import java.util.ArrayList;
import java.util.StringTokenizer;
import java.util.Properties;
import java.util.Enumeration;
import java.util.Collections;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;

import java.awt.Color;
import java.awt.Rectangle;
import java.awt.Component;

import java.net.URL;
import java.net.InetAddress;
import java.net.NetworkInterface;

/**
 * Dirk Krause's general tool class.
 * @author krause
 */
public class DkTool {

    /** Static strings. */
    protected static final String[] stt = {
        /*   0 */ "",
        /*   1 */ "\n",
        /*   2 */ " ",
        /*   3 */ "/",
        /*   4 */ "/all",
        /*   5 */ "true",
        /*   6 */ "false",
        /*   7 */ "http://",
        /*   8 */ "ui.languages",
        /*   9 */ "en",
        /*  10 */ "LANG",
        /*  11 */ "prefs.use-system",
        /*  12 */ " \t\r\n,;",
        /*  13 */ "\\",
    };

    /** Hexadecimal characters, used by encodeToHtml. */
    protected static final char[] hexCharacters = {
        '0', '1', '2', '3', '4', '5', '6', '7',
        '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
    };

    /** Strings which can be converted into boolean values. */
    protected static final String[] theBooleanStrings = {
        "y", "yes", "on", "true", "ok", "1", "+",
        "n", "no", "off", "false", "0", "-"
    };
    
    /** Index of first "no" entry in theBooleanStrings. */
    protected static final int first_no_boolean = 7;

    /** Allowed character set in PS files. */
    protected static final String psAllowedCharacters =
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789+-*/?.,;";

    /** Synchronization objects.
     *
     */
    private Object[] so = null;

    /** The language(s) for the user. */
    private String[] languages = null;

    /** Application name. */
    private String appName = null;

    /** Application group name (may be null). */
    private String groupName = null;

    /** Preferences nodes. */
    private Preferences[] myprefs = null;

    /** List of all resource base directories. */
    private ArrayList<String> resourceBaseList = null;

    /** Array of all resource base directories. */
    private String[] resourceBases = null;

    /** Flag: resourceBases is up to date. */
    private boolean resourceBasesUpToDate = true;

    /** Number of preference nodes. */
    private static final int nop = 6;

    /** Index of synchronization object for resource base array list. */
    private static final int SO_RB_ARRAY_LIST = nop;

    /** Index of synchronization object for resource bases. */
    private static final int SO_RESOURCE_BASES = SO_RB_ARRAY_LIST + 1;

    /** Index of synchronization object for resource base flag. */
    private static final int SO_FLAG_RESOURCE_BASES = SO_RESOURCE_BASES + 1;

    /** Number of synchronization objects. */
    private static final int nso = SO_FLAG_RESOURCE_BASES + 1;

    /** Platform information. */
    private PlatformInformation pfi = null;

    /**
     * Create DkTool object.
     * @param aName Application name (must be set).
     * @param gName Application group name (may be null, but not recommended).
     */
    public DkTool(String aName, String gName, PlatformInformation p) {
        int i;
        String l;
        String t;
        boolean cc;
        /* Set application and group name. */
        appName = aName; groupName = gName; pfi = p;
        /* Initialize synchronization objects. */
        so = new Object[nso];
        for(i = 0; i < nso; i++) { so[i] = new Object(); }
        /* Initialize preferences management. */
        myprefs = new Preferences[6];
        myprefs[0] = null;
        try {
            myprefs[0] = Preferences.userRoot().node(stt[3] + aName);
        } catch(Exception e) { myprefs[0] = null; }
        myprefs[1] = null;
        myprefs[2] = null;
        try {
            myprefs[2] = Preferences.userRoot().node(stt[3] + gName);
        } catch(Exception e) { myprefs[2] = null;}
        myprefs[3] = null;
        myprefs[4] = null;
        try {
            myprefs[4] = Preferences.userRoot().node(stt[4]);
        } catch(Exception e) { myprefs[4] = null; }
        myprefs[5] = null;
        if(getPreferenceBool(stt[11], false)) {
            try {
                myprefs[1] = Preferences.systemRoot().node(stt[3] + aName);
            } catch(Exception e) { myprefs[1] = null; }
            try {
                myprefs[3] = Preferences.systemRoot().node(stt[3] + gName);
            } catch(Exception e) { myprefs[3] = null; }
            try {
                myprefs[5] = Preferences.systemRoot().node(stt[4]);
            } catch(Exception e) { myprefs[5] = null; }
        }
        /* Set up the list of languages. */
        l = null;
        try {
            l = System.getenv(stt[10]);
	    int ind = l.indexOf('_');
	    if(ind > -1) {
	        String l1 = l.substring(0, ind);
		String l2 = l.substring(ind + 1);
		l = l1.toLowerCase() + "_" + l2.toUpperCase();
	    } else {
	        String l1 = l;
		l = l1.toLowerCase();
	    }
        } catch(Exception e) { l = null; }
        if(l == null) {
            try {
                Properties prop = System.getProperties();
                l = (String)prop.get("user.language");
                String ll = l.toLowerCase();
                l = ll;
                try {
                    String cou = (String)prop.get("user.country");
                    ll = cou.toUpperCase();
                    l = l + "_" + ll;
                } catch(Exception e) {}
            } catch(Exception e) { l = null; }
        }
        if(l == null) {
            if(pfi == null) {
                pfi = new PlatformInformation();
            }
            if(pfi != null) {
                l = pfi.getUserLanguage();
	        int ind = l.indexOf('_');
	        if(ind > -1) {
	            String l1 = l.substring(0, ind);
		    String l2 = l.substring(ind + 1);
		    l = l1.toLowerCase() + "_" + l2.toUpperCase();
	        } else {
	            String l1 = l;
		    l = l1.toLowerCase();
	        }
            }
        }
        if(l == null) {
            l = getPreferenceString(stt[8], stt[9]);
        }
        if(l != null) {
            ArrayList<String> als = new ArrayList<String>();
            Object[] oa;
            StringTokenizer stto = new StringTokenizer(l, stt[12]);
            boolean can_continue = true;
            while(stto.hasMoreTokens() && can_continue) {
                try {
                    t = stto.nextToken();
                    i = t.indexOf('.'); if(i >= 0) { t = t.substring(0, i); }
                    als.add(t);
                } catch(Exception e) { can_continue = false; }
            }
            oa = als.toArray();
            if(oa != null) {
                if(oa.length > 0) {
                    languages = new String[oa.length];
                    for(i = 0; i < oa.length; i++) {
                        languages[i] = new String((String)(oa[i]));
                    }
                }
            }
        }
    }

    /**
     * Constructor (simple version.
     * @param aName Applicatzion name.
     * @param gName Application group name.
     */
    public DkTool(String aName, String gName) {
        this(aName, gName, null);
    }



    /**
     * Retrieve preference string value.
     * @param k Preference key (name).
     * @param d Default value (may be null).
     * @return  Preference value or d.
     */
    public String getPreferenceString(String k, String d) {
        String back = null;
        for(int i = 0; ((i < 6) && (back == null)); i++) {
            if(myprefs[i] != null) {
                synchronized(so[i]) {
                    try {
                        back = myprefs[i].get(k, d);
                    } catch(Exception e) { back = null; }
                }
            }
        }
        return back;
    }
    
    
    /**
     * Retrieve int value from preferences.
     * @param k Preference key (name).
     * @param d Default value.
     * @return  The found value or the default value.
     */
    public int getPreferenceInt(String k, int d) {
        boolean found = false;
        int back;
        String s;
        int v;
        back = d;
        for(int i = 0; ((i < 6) && (!found)) ; i++) {
            if(myprefs[i] != null) {
                s = null;
                synchronized(so[i]) {
                    try {
                        s = myprefs[i].get(k, null);
                    } catch(Exception e) { s = null; }
                }
                if(s != null) {
                    try {
                        v = Integer.parseInt(s);
                        found = true;
                        back = v;
                    } catch(Exception e) {}
                }
            }
        }
        return back;
    }


    /**
     * Retrieve boolean value from preferences.
     * @param k Preference key.
     * @param d Default value.
     * @return  The value found or the default value d.
     */
    public boolean getPreferenceBool(String k, boolean d) {
        boolean found = false;
        boolean back;
        back = d;
        String s;
        for(int i = 0; ((i < myprefs.length) && (!found)); i++) {
            if(myprefs[i] != null) {
                s = null;
                synchronized(so[i]) {
                    try {
                        s = myprefs[i].get(k, null);
                    } catch(Exception e) { s = null; }
                }
                if(s != null) {
                    if(s.equalsIgnoreCase(stt[5])) {
                        back = true; found = true;
                    } else {
                        if(s.equalsIgnoreCase(stt[6])) {
                            back = false; found = true;
                        }
                    }
                }
            }
        }
        return back;
    }


    /**
     * Set boolean preference value.
     * @param k Preference ky.
     * @param v New value.
     * @param g Flag: Set value for application group too.
     * @param c Flag: Set value for common settings too.
     * @param s Flag: Set value in system nodes too.
     * @return  Flag to indicate success.
     */
    public boolean setPreferenceBool(
        String k, boolean v, boolean g, boolean c, boolean s
    )
    {
        boolean back = true;
        for(int i = 0; i < 3; i++) {
            int j = 2 * i;
            if((i == 0) || ((i == 1) && g) || ((i == 2) && c)) {
                if(myprefs[j] != null) {
                    synchronized(so[j]) {
                        try {
                            myprefs[j].put(k, (v ? stt[5] : stt[6]));
                        } catch(Exception e) { back = false; }
                    }
                }
                if(s) {
                    j++;
                    if(myprefs[j] != null) {
                        synchronized(so[j]) {
                            try {
                                myprefs[j].put(k, (v ? stt[5] : stt[6]));
                            } catch(Exception e) { back = false; }
                        }
                    }
                }
            }
        }
        return back;
    }



    /**
     * Set boolean preference value.
     * @param k Preference ky.
     * @param v New value.
     * @return  Flag to indicate success.
     */
    public boolean setPreferenceBool(String k, boolean v) {
        boolean back;
        back = setPreferenceBool(k, v, false, false, false);
        return back;
    }



    /**
     * Set preference value.
     * @param k Preference key (name).
     * @param v Value to set.
     * @param g Flag: set value in application group node too.
     * @param c Flag: set value in common preferences node too.
     * @param s Flag: Settings in the system nodes too.
     * @return  Flag to indicate success.
     */
    public boolean setPreferenceString(
        String k, String v, boolean g, boolean c, boolean s
    )
    {
        boolean back = true;
        for(int i = 0; i < 3; i++) {
            int j = 2 * i;
            if((i == 0) || ((i == 1) && g) || ((i == 2) && c)) {
                if(myprefs[j] != null) {
                    synchronized(so[j]) {
                        try {
                            myprefs[j].put(k, ((v != null) ? v : stt[0]));
                        } catch(Exception e) { back = false; }
                    }
                }
                if(s) {
                    j++;
                    if(myprefs[j] != null) {
                        synchronized(so[j]) {
                            try {
                                myprefs[j].put(k, ((v != null) ? v : stt[0]));
                            } catch(Exception e) { back = false; }
                        }
                    }
                }
            }
        }
        return back;
    }


    /**
     * Set int preference value.
     * @param k Preference key (name).
     * @param v Value to set.
     * @param g Flag: Set entry for application group too.
     * @param c Flag: Set entry for all applications too.
     * @param s Flag: Attempt to make settings in system nodes too.
     * @return  Flag to indicate success.
     */
    public
    boolean
    setPreferenceInt(String k, int v, boolean g, boolean c, boolean s)
    {
        boolean back = true;
        String x = (new Integer(v)).toString(); // String.format("%d", v);
        for(int i = 0; i < 3; i++) {
            int j = 2 * i;
            if((i == 0) || ((i == 1) && g) || ((i == 2) && c)) {
                if(myprefs[j] != null) {
                    synchronized(so[j]) {
                        try {
                            myprefs[j].put(k, x);
                        } catch(Exception e) { back = false; }
                    }
                }
                if(s) {
                    j++;
                    if(myprefs[j] != null) {
                        synchronized(so[j]) {
                            try {
                                myprefs[j].put(k, x);
                            } catch(Exception e) { back = false; }
                        }
                    }
                }
            }
        }
        return back;
    }


    /**
     * Set int preference value.
     * @param k Preference key (name).
     * @param v Value to set.
     * @param g Flag: Set entry for application group too.
     * @param c Flag: Set entry for all applications too.
     * @return  Flag to indicate success.
     */
    public
    boolean
    setPreferenceInt(String k, int v, boolean g, boolean c)
    {
        boolean back;
        back = setPreferenceInt(k, v, g, c, false);
        return back;
    }

    
    
    /**
     * Set int preference value.
     * @param k Preference key (name).
     * @param v Value to set.
     * @return  Flag to indicate success.
     */
    public
    boolean
    setPreferenceInt(String k, int v)
    {
        boolean back;
        back = setPreferenceInt(k, v, false, false, false);
        return back;
    }


    /**
     * Set preference value.
     * @param k Preference key (name).
     * @param v Value to set.
     * @return  Flag to indicate success.
     */
    public boolean setPreferenceString(String k, String v, boolean g, boolean c)
    {
        boolean back = true;
        back = setPreferenceString(k, v, g, c, false);
        return back;
    }


    /**
     * Set preference value.
     * @param k Preference key (name).
     * @param v New value.
     * @return  Flag to indicate success.
     */
    public boolean setPreferenceString(String k, String v) {
        boolean back = setPreferenceString(k, v, false, false, false);
        return back;
    }



    /**
     * Flush all preferences used by this object.
     * @return  Flag to indicate success.
     */
    public boolean flushPreferences() {
        boolean back = true;
        for(int i = 0; i < 3; i++) {
            int j = 2 * i;
            if(myprefs[j] != null) {
                synchronized(so[j]) {
                    try {
                        myprefs[j].sync();
                    } catch(Exception e) { back = false; }
                }
            }
        }
        return back;
    }



    /**
     * Bulk-retrieval of preference values.
     * @param k Array of preference keys.
     * @param v Result array (input: default values, output: found values).
     */
    public void getMultiplePreferenceStrings(String[] k, String[] v) {
        int max = k.length; /* Maximum number of elements in arrays. */
        int i;              /* Current element in array. */
        int p;              /* Current preference node. */
        boolean done;       /* Current element in array is finished. */
        String value;       /* Value for current element. */
        if(max > v.length) { max = v.length; }
        for(i = 0; i < max; i++) {
            done = false;
            for(p = 0; ((p < 6) && (!done)); p++) {
                if(myprefs[p] != null) {
                    value = null;
                    synchronized(so[p]) {
                        try {
                            value = myprefs[p].get(k[i], null);
                        } catch(Exception e) { done = false; value = null; }
                        if(value != null) {
                            v[i] = value; done = true;
                        }
                    }
                }
            }
        }
    }



    /**
     * Bulk-retrieval of boolean preferences.
     * @param k Preference keys array.
     * @param v Boolean value array already containing defaults.
     */
    public void getMultiplePreferenceBools(String[] k, boolean[] v) {
        int max;
        int i;
        int p;
        boolean found;
        String s;
        max = k.length;
        if(max > v.length) { max = v.length; }
        for(i = 0; i < max; i++) {
            found = false;
            for(p = 0; ((p < myprefs.length) && (!found)); p++) {
                if(myprefs[p] != null) {
                    s = null;
                    synchronized(so[p]) {
                        try {
                            s = myprefs[p].get(k[i], null);
                        } catch(Exception e) { s = null; }
                    }
                    if(s != null) {
                        if(s.equalsIgnoreCase(stt[5])) {
                            v[i] = true; found = true;
                        } else {
                            if(s.equalsIgnoreCase(stt[6])) {
                                v[i] = false; found = true;
                            }
                        }
                    }
                }
            }
        }
    }


    /**
     * Retrieve multiple int values from preferences.
     * @param k Keys array.
     * @param v Values array already containing the default values.
     */
    public void getMultiplePreferenceInts(String[] k, int[] v) {
        int max;
        int i;
        int p;
        int t;
        boolean done;
        String value;
        max = k.length;
        if(max > v.length) { max = v.length; }
        for(i = 0; i < max; i++) {
            done = false;
            for(p = 0; ((p < 6) && (!done)); p++) {
                if(myprefs[p] != null) {
                    value = null;
                    synchronized(so[p]) {
                        try {
                            value = myprefs[p].get(k[i], null);
                        } catch(Exception e) { value = null; done = false; }
                    }
                    if(value != null) {
                        try {
                            t = Integer.parseInt(value);
                            done = true;
                            v[i] = t;
                        } catch(Exception e) { done = false; }
                    }
                }
            }
        }
    }


    /**
     * Set multiple int preference values.
     * @param k Keys array.
     * @param v Values array.
     * @param g Flag: Set values for the application group.
     * @param c Flag: Set values for all applications.
     * @param s Flag: Attempt settings in system nodes too.
     * @return  Flag to indicate success.
     */
    public boolean setMultiplePreferenceInts(
        String[] k, int[] v, boolean g, boolean c, boolean s
    )
    {
        boolean back = true;
        int i;
        int j;
        int p;
        int max;
        String value;
        max = k.length;
        if(max > v.length) { max = v.length; }
        for(i = 0; i < max; i++) {
            // value = String.format("%d", v[i]);
	    value = (new Integer(v[i])).toString();
            for(j = 0; j < 3; j++) {
                p = 2 * j;
                if((j == 0) || ((j == 1) && g) || ((j == 2) && c)) {
                    if(myprefs[p] != null) {
                        synchronized(so[p]) {
                            try {
                                myprefs[p].put(k[i], value);
                            } catch(Exception e) { back = false; }
                        }
                    }
                    if(s) {
                        p++;
                        if(myprefs[p] != null) {
                            synchronized(so[p]) {
                                try {
                                    myprefs[p].put(k[i], value);
                                } catch(Exception e) { back = false; }
                            }
                        }
                    }
                }
            }
        }
        return back;
    }


    /**
     * Set multiple int preference values.
     * @param k Keys array.
     * @param v Values array.
     * @return  Flag to indicate success.
     */
    public boolean setMultiplePreferenceInts(String[] k, int[] v) {
        boolean back;
        back = setMultiplePreferenceInts(k, v, false, false, false);
        return back;
    }


    /**
     * Bulk-set of preference values.
     * @param k Array of preferences keys (names).
     * @param v Array of preferences values.
     * @param g Flag: Set entries in application group node too.
     * @param c Flag: Set entries in common node too.
     * @param s Flag: Attempt settings in the system nodes too.
     * @return  Flag to indicate success.
     */
    public boolean setMultiplePreferenceStrings(
        String[] k, String[] v, boolean g, boolean c, boolean s
    )
    {
        boolean back = true;
        int max = k.length;
        int i;
        int j;
        int p;
        if(max > v.length) { max = v.length; }
        for(i = 0; i < max; i++) {
            for(j = 0; j < 3; j++) {
                p = 2 * j;
                if((j == 0) || ((j == 1) && g) || ((j == 2) && c)) {
                    if(myprefs[p] != null) {
                        synchronized(so[p]) {
                            try {
                                myprefs[p].put(
                                    k[i], ((v[i] != null) ? v[i] : stt[0])
                                );
                            } catch(Exception e) { back = false; }
                        }
                    }
                    if(s) {
                        p++;
                        if(myprefs[p] != null) {
                            synchronized(so[p]) {
                                try {
                                    myprefs[p].put(
                                        k[i], ((v[i] != null) ? v[i] : stt[0])
                                    );
                                } catch(Exception e) { back = false; }
                            }
                        }
                    }
                }
            }
        }
        return back;
    }


    /**
     * Bulk-set of preference values.
     * @param k Array of preferences keys (names).
     * @param v Array of preferences values.
     * @return  Flag to indicate success.
     */
    public boolean setMultiplePreferenceStrings(String[] k, String[] v) {
        boolean back;
        back = setMultiplePreferenceStrings(k, v, false, false, false);
        return back;
    }


    /**
     * Bulk-set of boolean preference values.
     * @param k Array of preference keys (names).
     * @param v Array of preference values.
     * @param g Flag: Apply settings to application group.
     * @param c Flag: Apply settings as general defaults.
     * @param s Flag: Apply settings to system nodes too.
     * @return  Flag to indicate success.
     */
    public boolean setMultiplePreferenceBools(
        String[] k, boolean[] v, boolean g, boolean c, boolean s
    )
    {
        boolean back = true;
        int max;
        int i;
        int p;
        int j;
        String str;
        max = k.length;
        if(max > v.length) { max = v.length; }
        for(i = 0; i < max; i++) {
            str = (v[i] ? stt[5] : stt[6]);
            for(j = 0; j < 3; j++) {
                if((j == 0) || ((j == 1) && g) || ((j == 2) && c)) {
                    p = 2 * j;
                    if(myprefs[p] != null) {
                        synchronized(so[p]) {
                            try {
                                myprefs[p].put(k[i], str);
                            } catch(Exception e) { back = false; }
                        }
                    }
                    if(s) {
                        p++;
                        if(myprefs[p] != null) {
                            synchronized(so[p]) {
                                try {
                                    myprefs[p].put(k[i], str);
                                } catch(Exception e) { back = false; }
                            }
                        }
                    }
                }
            }
        }
        return back;
    }


    /**
     * Bulk-set of boolean preference values.
     * @param k Array of preference keys (names).
     * @param v Array of preference values.
     * @return  Flag to indicate success.
     */
    public boolean setMultiplePreferenceBools(String[] k, boolean[] v) {
        boolean back;
        back = setMultiplePreferenceBools(k, v, false, false, false);
        return back;
    }

    /**
     * Add a resource base (synchronized).
     * @param s New base directory to add.
     */
    public void addResourceBase(String s) {
        synchronized(so[SO_RB_ARRAY_LIST]) {
            resourceBaseList.add(s);
        }
        synchronized(so[SO_FLAG_RESOURCE_BASES]) {
            resourceBasesUpToDate = false;
        }
    }



    /**
     * Rebuild the resource bases string array if necessary.
     */
    private void rebuildResourceBasesIfNecessary() {
        boolean mustRebuild = false;
        String[] newArray = null;
        synchronized(so[SO_FLAG_RESOURCE_BASES]) {
            mustRebuild = resourceBasesUpToDate;
        }
        mustRebuild = !mustRebuild;
        if(mustRebuild) {
            Object[] oa = null;
            synchronized(so[SO_RB_ARRAY_LIST]) {
                oa = resourceBaseList.toArray();
            }
            if(oa != null) {
                newArray = new String[oa.length];
                for(int i = 0; i < oa.length; i++) {
                    newArray[i] = new String((String)(oa[i]));
                }
            }
            synchronized(so[SO_RESOURCE_BASES]) {
                resourceBases = newArray;
            }
            synchronized(so[SO_FLAG_RESOURCE_BASES]) {
                resourceBasesUpToDate = true;
            }
        }
    }



    /**
     * Get number of entries in resource base array.
     * @return  Number of entries in resource base array.
     */
    public int getResourceBaseNumber() {
        int back = 0;
        rebuildResourceBasesIfNecessary();
        synchronized(so[SO_RESOURCE_BASES]) {
            if(resourceBases != null) {
                back = resourceBases.length;
            }
        }
        return back;
    }



    /**
     * Get one entry from the resouce base array.
     * @param i Index of entry in array.
     * @return  Entry from resource base array or null.
     */
    public String getResourceBase(int i) {
        String back = null;
        rebuildResourceBasesIfNecessary();
        synchronized(so[SO_RESOURCE_BASES]) {
            if(resourceBases != null) {
                if((i >= 0) && (i < resourceBases.length)) {
                    back = resourceBases[i];
                }
            }
        }
        if(back != null) { back = new String(back); }
        return back;
    }



    /**
     * Open a resource (URL or file in jar archive)  and return input stream.
     * @param o     Object requiring the resource.
     * @param fn    File name or URL of resource.
     * @return      Input stream to read resource or null.
     */
    public InputStream openResource(Object o, String fn) {
        InputStream back = null;
        URL url = null;
        boolean have_slash = false;
        String rn;
	
        if(fn.startsWith(stt[7])) { /* http:// (full URL specified). */
            try {
                url = new URL(fn);
            } catch(Exception e) { url = null; }
            if(url != null) {
                try {
                    back = url.openStream();
                } catch(Exception e) { back = null; }
            }
        } else {			
            if(fn.startsWith(stt[3])) { /* Full file name specified. */
                have_slash = true;
                try {			
                    back = o.getClass().getResourceAsStream(fn);
                } catch(Exception e) { back = null; }
            } else {
                try {			
                    back = o.getClass().getResourceAsStream("/" + fn);
                } catch(Exception e) { back = null; }
            }
            if(back == null) {  /* We must walk through the resource bases. */
                String[] rb = null;
                String d;		
                rebuildResourceBasesIfNecessary();
                synchronized(so[SO_RESOURCE_BASES]) {
                    rb = resourceBases;
                }
                if(rb != null) {
                    for(int i = 0; ((i < rb.length) && (back == null)); i++) {
                        synchronized(so[SO_RESOURCE_BASES]) {
                            d = rb[i];
                        }
                        if(d != null) {
                            if(have_slash) {
                                rn = d + fn;
                            } else {
                                rn = d + "/" + fn;
                            } 
                            if(rn.startsWith(stt[7])) {
                                try {
                                    url = new URL(rn);
                                } catch(Exception e) { url = null; }
                                if(url != null) {
                                    try {
                                        back = url.openStream();
                                    } catch(Exception e) { back = null; }
                                }
                            } else {
                                if(rn.startsWith(stt[3])) {
                                    try {
                                        back = o.getClass().getResourceAsStream(
                                            rn
                                        );
                                    } catch(Exception e) { back = null; }
                                } else {
                                    try {
                                        back = o.getClass().getResourceAsStream(
                                            rn
                                        );
                                    } catch(Exception e) { back = null; }
                                }
                            }
                        }
                    }
                }
            } else {		
	    }
        } 
        return back;
    }


    /**
     * Open a localized resource.
     * @param o     Object requiring the resource.
     * @param fn    File name or URL.
     * @return      InputStream on success, null on error.
     */
    public InputStream openLocalizedResource(Object o, String fn) {
        InputStream back = null;
        int l;
        int i;
        int passno;
        String la;
        String re;
        String newfn;
        StringBuilder sb;
	
        boolean has_slash = false;
        if(fn.startsWith(stt[3])) {		
	    has_slash = true;
	}
        if(languages != null) {			
            if(languages.length > 0) {		
                for(l = 0; ((l < languages.length) && (back == null)); l++) {
                    la = languages[l]; re = null;	
                    i = la.indexOf('_');
                    if(i >= 0) {
                        re = la.substring(i + 1);
                        la = la.substring(0, i); 
                    }
                    passno = 1;
                    if(re != null) {
                        passno = 2;
                        re = re.toLowerCase();
                    }
                    while((passno >= 0) && (back == null)) {
                        sb = new StringBuilder();
                        switch(passno) {
                            case 2: {
                                sb.append(stt[3]);
                                sb.append(la);
                                sb.append(stt[3]);
                                sb.append(re);
                                if(!has_slash) { sb.append(stt[3]); }
                                sb.append(fn);
                            } break;
                            case 1: {
                                sb.append(stt[3]);
                                sb.append(la);
                                if(!has_slash) { sb.append(stt[3]); }
                                sb.append(fn);
                            } break;
                            default: {
                                if(!has_slash) { sb.append(stt[3]); }
                                sb.append(fn);
                            } break;
                        }
                        newfn = new String(sb);		
                        back = openResource(o, newfn);
			if(back != null) {		
			}
                        passno--;
                    }
                }
            } else {			
                back = openResource(o, fn);
            }
        } else {			
            back = openResource(o, fn);
        } 
        return back;
    }

    /**
     * Corect one line read from a localized texts file.
     * @param s Original string.
     * @return  String after processing backslashes
     */
    public static String processBackslashEscapedString(String s) {
        String back = null;
        StringBuilder sb = new StringBuilder();
        int l = s.length();
        char c;
        int state = 0;
        for(int i = 0; i < l; i++) {
            c = s.charAt(i);
            if(state == 1) {
                switch(c) {
                    case '"': {
                        sb.append('"');
                    } break;
                    case 'n': {
                        sb.append('\n');
                    } break;
                    case 'r': {
                        sb.append('\r');
                    } break;
                    case 't': {
                        sb.append('\t');
                    } break;
                    case 'b': {
                        sb.append('\b');
                    } break;
                    default: {
                        sb.append(c);
                    } break;
                }
                state = 0;
            } else {
                if(c == '\\') {
                    state = 1;
                } else {
                    sb.append(c);
                }
            }
        }
        back = new String(sb);
        if(back == null) { back = stt[0]; }
        return back;
    }

    /**
     * Construct one large string from an input stream contents.
     * @param is    Input stream to process.
     * @param csn   Character set name ("UTF-8" or "ISO-8859-1" or null).
     * @return      A string containing the file contents.
     */
    public String getStringFromInputStream(InputStream is, String csn) {
        String back = null;
        StringBuilder sb = new StringBuilder();
        InputStreamReader isr = null;
        BufferedReader br = null;
        try {
            if(csn != null) { isr = new InputStreamReader(is, csn); }
            else { isr = new InputStreamReader(is); }
            br = new BufferedReader(isr);
            String lineBuffer;
            boolean cc = true;
            while(cc) {
                lineBuffer = null; cc = false;
                try {
                    lineBuffer = br.readLine();
                } catch(Exception e) { lineBuffer = null; }
                if(lineBuffer != null) {
                    cc = true;
                    sb.append(lineBuffer);
                }
            }
        }
        catch(Exception e) {
        }
        finally {
            try { br.close(); } catch(Exception e) {}
            try { isr.close(); } catch(Exception e) {}
        }
        back = new String(sb);
        return back;
    }


    /**
     * Construct one large string from an input stream contents.
     * @param is    Input stream to process.
     * @return      A string containing the file contents.
     */
    public String getStringFromInputStream(InputStream is) {
        String back = getStringFromInputStream(is, null);
        return back;
    }


    /**
     * Read array of strings from an input stream.
     * @param is    Input stream.
     * @param csn   Character set name ("UTF-8" or "ISO-8859-1").
     * @return      String array or null.
     */
    public String[] getArrayFromInputStream(InputStream is, String csn) {
        String[] back = null;
        InputStreamReader isr = null;
        BufferedReader br = null;
        ArrayList<String> al = new ArrayList<String>();
        try {
            if(csn != null) { isr = new InputStreamReader(is, csn); }
            else { isr = new InputStreamReader(is); }
            br = new BufferedReader(isr);
            String lineBuffer;
            boolean cc;
            cc = true;
            while(cc) {
                lineBuffer = null; cc = false;
                try {
                    lineBuffer = br.readLine();
                }
                catch(Exception e) {
                    lineBuffer = null;
                }
                if(lineBuffer != null) {
                    cc = true;
                    if(!(lineBuffer.startsWith("#"))) {
                        lineBuffer = processBackslashEscapedString(lineBuffer);
                        al.add(lineBuffer);
                    }
                }
            }
            Object[] objarr = al.toArray();
            back = new String[objarr.length];
            for(int i = 0; i < objarr.length; i++) {
                back[i] = new String((String)(objarr[i]));
            }
        }
        catch(Exception e) {
            back = null;
        }
        finally {
            try { br.close(); } catch(Exception e) {}
            try { isr.close(); } catch(Exception e) {}
        }
        return back;
    }


    /**
     * Read array of strings from an input stream.
     * @param is    Input stream.
     * @return      String array or null.
     */
    public String[] getArrayFromInputStream(InputStream is) {
        String[] back = null;
        back = getArrayFromInputStream(is, null);
        return back;
    }


    /**
     * Retrieve array of localized strings from file.
     * @param o     Object requiring the strings.
     * @param fn    File name.
     * @param csn   Character set name ("UTF-8" or "ISO-8859-1").
     * @return      String array or null.
     */
    public String[] getLocalizedArray(Object o, String fn, String csn) {
        String[] back = null;
        InputStream is = null;
        is = openLocalizedResource(o, fn);
        if(is != null) {
            back = getArrayFromInputStream(is, csn);
            try { is.close(); } catch(Exception e) {}
        }
        return back;
    }



    /**
     * Retrieve array of localized strings from file.
     * @param o     Object requiring the strings.
     * @param fn    File name.
     * @return      String array or null.
     */
    public String[] getLocalizedArray(Object o, String fn) {
        String[] back = null;
        back = getLocalizedArray(o, fn, null);
        return back;
    }


    /**
     * Get file contents as one string.
     * @param o     Object.
     * @param fn    File name.
     * @param csn   Character set name.
     * @return      File contents as one large string.
     */
    public String getLocalizedString(Object o, String fn, String csn) {
        String back = null;
        InputStream is = null;
        is = openLocalizedResource(o, fn);
        if(is != null) {
            back = getStringFromInputStream(is, csn);
            try { is.close(); } catch(Exception e) {}
        }
        return back;
    }


    /**
     * Get file contents as one string.
     * @param o     Object.
     * @param fn    File name.
     * @return      File contents as one large string.
     */
    public String getLocalizedString(Object o, String fn) {
        String back = getLocalizedString(o, fn, null);
        return back;
    }



    /**
     * Retrieve integer value from string.
     * @param s String to inspect.
     * @param d Default value.
     * @return  Conversion result or default value.
     */
    public static int getIntValue(String s, int d) {
        int back = d;
        int base = 10;
        if(s.startsWith("0x")) {
            base = 16;
        } else {
            if(s.startsWith("0")) {
                base = 8;
            }
        }
        try {
            back = Integer.parseInt(s, base);
        } catch(Exception e) { back = d; }
        return back;
    }


    /**
     * Retrieve integer value from string.
     * @param s String to inspect.
     * @return  Conversion result or 0 if conversion failed.
     */
    public static int getIntValue(String s) {
        int back = getIntValue(s, 0);
        return back;
    }



    /**
     * Convert a string into a color.
     * @param s     String to convert.
     * @param dr    Default red.
     * @param dg    Default green.
     * @param db    Default blue.
     * @return      The color retrieved from the string.
     */
    public static Color convertColor(String s, int dr, int dg, int db) {
        int r = dr; int g = dg; int b = db;
        String x;
        if(s != null) {
            try {
                StringTokenizer st = new StringTokenizer(s, ",");
                if(st.hasMoreTokens()) {
                    x = st.nextToken();
                    r = getIntValue(x, dr);
                    if(st.hasMoreTokens()) {
                        x = st.nextToken();
                        g = getIntValue(x, dg);
                        if(st.hasMoreTokens()) {
                            x = st.nextToken();
                            b = getIntValue(x, db);
                        }
                    }
                }
            } catch(Exception e) { r = dr; g = dg; b = db; }
        }
        return(new Color(r, g, b));
    }


    /**
     * Convert a string into a color.
     * @param s     String to convert.
     * @return      The color retrieved from the string.
     */
    public static Color convertColor(String s) {
        Color back = convertColor(s, 255, 255, 255);
        return back;
    }


    /**
     * Get index of string s in string array a.
     * @param a     Array with patterns.
     * @param s     String to search in array.
     * @param cs    Flag: Case-sensitive or not.
     * @return      Index of matched string on success, -1 on error.
     */
    public static int getStringArrayIndex(String[] a, String s, boolean cs) {
        int back = -1;
        int i;
        for(i = 0; ((i < a.length) && (back == -1)); i++) {
            if(cs) {
                if(s.equals(a[i])) { back = i; }
            } else {
                if(s.equalsIgnoreCase(a[i])) { back = i; }
            }
        }
        return back;
    }



    /**
     * Get index of string s in string array a.
     * @param a     Array to search.
     * @param s     String to search for.
     * @param c     Array of indexes to check.
     * @param cs    Flag: Case sensitive or not.
     * @return      Index in array or -1.
     */
    public static int getStringArrayIndex(
        String[] a, String s, int[] c, boolean cs)
    {
        int back = -1;
        int i;
        int ind;
        if(c != null) {
            if(c.length > 0) {
                for(i = 0; ((i < c.length) && (back == -1)); i++) {
                    ind = c[i];
                    if((ind >= 0) && (ind < a.length)) {
                        try {
                            if(cs) {
                                if(s.equals(a[ind])) {
                                    back = ind;
                                }
                            } else {
                                if(s.equalsIgnoreCase(a[ind])) {
                                    back = ind;
                                }
                            }
                        } catch(Exception e) {}
                    }
                }
            } else {
                back = getStringArrayIndex(a, s, cs);
            }
        } else {
            back = getStringArrayIndex(a, s, cs);
        }
        return back;
    }



    /**
     * Get index of string s in string array a.
     * @param a     Array to search.
     * @param s     String to search for.
     * @param c     Array of indexes to check.
     * @return      Index in array or -1.
     */
    public static int getStringArrayIndex(String[] a, String s, int[] c) {
        int back = getStringArrayIndex(a, s, c, true);
        return back;

    }


    /**
     * Get index of string s in string array a.
     * @param a Array with patterns.
     * @param s String to search in array.
     * @return  Index of matched string on success, -1 on error.
     */
    public static int getStringArrayIndex(String[] a, String s) {
        int back = getStringArrayIndex(a, s, true);
        return back;
    }


    /**
     * Convert string to boolean.
     * @param s String to convert.
     * @param d Default value.
     * @return  Conversion result or default value.
     */
    public static boolean getBoolean(String s, boolean d) {
        boolean back = d;
        int i = getStringArrayIndex(theBooleanStrings, s);
        if(i > -1) {
            if(i < first_no_boolean) back = true;
            else back = false;
        }
        return back;
    }



    /**
     * Check whether a string is a boolean.
     * @param s String to check.
     * @return  Check result.
     */
    public static boolean isBoolean(String s) {
        boolean back = false;
        if(getStringArrayIndex(theBooleanStrings, s) > -1) {
            back = true;
        }
        return back;
    }


    /**
     * Convert a text string to PS encoding, so we can insert it into
     * a PS file.
     * @param s String to convert.
     * @return  Conversion result.
     */
    public static String encodeForPostscript(String s) {
      String back;
      StringBuilder res = new StringBuilder();
      int i;
      char c;
      int ic;
      String x;
      for(i = 0; i < s.length(); i++) {
          c = s.charAt(i);
          if(psAllowedCharacters.indexOf((int)c) >= 0) {
              res.append(c);
          } else {
              switch(c) {
                  case '\n': { res.append("\\n"); } break;
                  case '\t': { res.append("\\t"); } break;
                  case '\f': { res.append("\\f"); } break;
                  case '\r': { res.append("\\r"); } break;
                  case '\b': { res.append("\\b"); } break;
                  case '\\': { res.append("\\\\"); } break;
                  default: {
                      ic = (int)c;
                      if((ic >= 0) && (ic <= 255)) {
                          x = Integer.toOctalString(ic);
                          switch(x.length()) {
                              case 0: { res.append("\\000"); } break;
                              case 1: {
                                  res.append("\\00"); res.append(x);
                              } break;
                              case 2: {
                                  res.append("\\0"); res.append(x);
                              } break;
                              case 3: {
                                  res.append("\\"); res.append(x);
                              } break;
                          }
                      }
                  } break;
              }
          }
      }
      back = new String(res);
      return back;
  }

    /**
     * Choose value for one co-ordinate.
     *
     * @param sp
     * Start value of parent component.
     *
     * @param wp
     * Width value of parent component.
     *
     * @param wc
     * Width value of child component.
     *
     * @param wd
     * Width value of screen.
     *
     * @return
     * A start value for the child component.
     */
    private static int cp(int sp, int wp, int wc, int wd)
    {
        int back = 0;
        if((sp + wc) <= wd) {
            back = sp;
        } else {
            if((sp + wp) >= wc) {
                back = sp + wp - wc;
                // back = sp + wp - wc + 1;
            } else {
                back = sp + ((wp - wc) / 2);
            }
        }
        return back;
    }

    /**
     * Correct value for one co-ordinate
     *
     * @param x
     * The original (uncorrected) value.
     *
     * @param cw
     * Component width in this direction.
     *
     * @param dw
     * The screen width in this direction.
     *
     * @return
     * The corrected value.
     */
    private static int cor(int x, int cw, int dw)
    {
        int back = x;
        if((x + cw) > dw) {
            back = dw - cw;
        }
        if(back < 0) {
            back = 0;
        }
        return back;
    }

  /**
   * Choose position for component c relative to parent component p.
   *
   * @param w
   * Screen width.
   *
   * @param h
   * Screen height.
   *
   * @param p
   * Parent component.
   *
   * @param c
   * New child component to be placed.
   *
   * @param mod
   * Flag to indicate whether or not c is a modal component.
   */
    public static void choosePosition(
        int w, int h, Component p, Component c, boolean mod
    )
    {
        int x = -1;
        int y = -1;
        Rectangle pr = new Rectangle();
        Rectangle cr = new Rectangle();
        p.getBounds(pr);
        c.getBounds(cr);
        boolean b = false;
        if(!mod) {
            // right
            if((pr.x + pr.width + cr.width) < w) {
                x = pr.x + pr.width;
                y = cp(pr.y, pr.height, cr.height, h);
                b = true;
            }
            // left
            if(!b) {
                if(cr.width < pr.x) {
                    x = pr.x - cr.width;
                    y = cp(pr.y, pr.height, cr.height, h);
                    b = true;
                }
            }
            // bottom
            if(!b) {
                if((pr.y + pr.height + cr.height) < h) {
                    y = pr.y + pr.height;
                    x = cp(pr.x, pr.width, cr.width, w);
                    b = true;
                }
            }
            // top
            if(!b) {
                if(cr.height < pr.y) {
                    y = pr.y - cr.height;
                    x = cp(pr.x, pr.width, cr.width, w);
                    b = true;
                }
            }
        }
        if(!b) {
            // above
            x = pr.x + ((pr.width - cr.width) / 2);
            y = pr.y + ((pr.height - cr.height) / 2);
        }
        x = cor(x, cr.width, w);
        y = cor(y, cr.height, h);
        c.setLocation(x, y);
    }

    /**
     * Choose modal position for component c relative to parent component p.
     *
     * @param w
     * Screen width.
     *
     * @param h
     * Screen height.
     *
     * @param p
     * Parent component.
     *
     * @param c
     * New modal child component to be placed.
     *
     */
    public static void chooseModalPosition(
        int w, int h, Component p, Component c
    )
    {
        choosePosition(w, h, p, c, true);
    }

    /**
     * Choose normal (non-modal) position for component c relative to
     * parent component p.
     *
     * @param w
     * Screen width.
     *
     * @param h
     * Screen height.
     *
     * @param p
     * Parent component.
     *
     * @param c
     * New non-modal child component to be placed.
     */
    public static void chooseNonModalPosition(
        int w, int h, Component p, Component c
    )
    {
        choosePosition(w, h, p, c, false);
    }


    /**
     * Correct position after resizing.
     *
     * @param w
     * Screen width.
     *
     * @param h
     * Screen height.
     * 
     * @param c
     * The component for which the position must be corrected.
     */
    public static void correctPosition(int w, int h, Component c) {
        Rectangle cr = new Rectangle();
        c.getBounds(cr);
        int x = cor(cr.x, cr.width, w);
        int y = cor(cr.y, cr.height, h);
        if((x != cr.x) || (y != cr.y)) {
            c.setLocation(x, y);
        }
    }


    /**
     * Get the users language.
     * @return  Language name or null.
     */
    public String getLanguage() {
        String back = null;
        if(languages != null) {
            if(languages.length > 0) {
                back = languages[0];
            }
        }
        if(back == null) { back = stt[9]; }
        back = new String(back);
        return back;
    }


    /**
     * Retrieve platform information object or create a new one.
     * @return  Platform information object.
     */
    public PlatformInformation getPfi() {
        PlatformInformation back = null;
        synchronized(this) {
            back = pfi;
        }
        if(back == null) {
            back = new PlatformInformation();
            synchronized(this) {
                pfi = back;
            }
        }
        return back;
    }

    /**
     * Get path separator string.
     * @return  Path separator string or null.
     */
    public String getPathSeparator() {
        String back = null;
        Properties prop = System.getProperties();
        try {
            back = (String)prop.get("file.separator");
        } catch(Exception e) { back = null; }
        if(back == null) {
            PlatformInformation p = getPfi();
            back = stt[p.getFilenamesSeparatorBackslash() ? 13 : 3];
        }
        return back;
    }



    /**
     * Get user name.
     * @return  User name or null.
     */
    public String getUserName() {
        String back = null;
        Properties prop = System.getProperties();
        try {
            back = (String)prop.get("user.name");
        } catch(Exception e) { back = null; }
        if(back == null) {
            PlatformInformation p = getPfi();
            back = p.getUserName();
        }
        if(back == null) {
            try {
                back = System.getenv("LOGNAME");
            } catch(Exception e) { back = null; }
        }
        if(back == null) {
            try {
                back = System.getenv("USERNAME");
            } catch(Exception e) { back = null; }
        }
        return back;
    }


    /**	Retrieve local host name (must appear in LPD request).
  	@return	Host name or null.
  */
    public static String getHostName() {
        String back = null;
        int level = 0;
        try {
            Enumeration<NetworkInterface> netInter =
            NetworkInterface.getNetworkInterfaces();
            while(netInter.hasMoreElements()) {
                NetworkInterface ni = netInter.nextElement();
                for(InetAddress ia : Collections.list(ni.getInetAddresses())) {
                    if(ia.isLoopbackAddress()) {
                        if(level == 0) {
                            back = ia.getHostAddress();
                            level = 1;
                        }
                    } else {
                        if(ia.isSiteLocalAddress()) {
                            if(level <= 1) {
                                back = ia.getHostAddress();
                                level = 2;
                            }
                        } else {
                            if(level <= 2) {
                                back = ia.getHostAddress();
                                level = 3;
                            }
                        }
                    }
                }
            }
        } catch(Exception e) {}
        return back;
    }


    /**
     * Find position of file name suffix.
     * @param fn    File name
     * @return Index of the suffix dot or -1.
     */
    public int getFilenameSuffixPosition(String fn) {
        int back = -1;
        char c;
        int l = fn.length();
        boolean sepbs = pfi.getFilenamesSeparatorBackslash();
        for(int i = 0; i < l; i++) {
            c = fn.charAt(i);
            if(c == '.') {
                back = i;
            } else {
                if(sepbs) {
                    if(c == '\\') { back = -1; }
                } else {
                    if(c == '/') { back = -1; }
                }
            }
        }
        return back;
    }


    /**
     * Find file name suffix (including the dot).
     * @param fn    File name to inspect.
     * @return  File name suffix or null.
     */
    public String getFilenameSuffix(String fn) {
        String back = null;
        int posdot = getFilenameSuffixPosition(fn);
        if(posdot >= 0) {
            back = fn.substring(posdot);
        }
        return back;
    }


    /**
     * Replace file suffixes by a new suffix (or add new suffix).
     * @param fn    Original file name.
     * @param ns    New suffix to append.
     * @return      The modified file name
     */
    public String replaceSuffix(String fn, String ns) {
        String back = null;
        int sp = getFilenameSuffixPosition(fn);
        if(sp > -1) {
            String bn = fn.substring(0, sp);
            back = new String(bn + ns);
        } else {
            back = new String(fn + ns);
        }
        return back;
    }

    /**
     * Return string for HTML encoding of one character.
     * @param c Character to encode.
     * @return  HTML representation of c.
     */
    private String htmlEncodedCharacter(char c) {
        StringBuilder sb = new StringBuilder();
        sb.append("&#x");
        if(c > 0xFFFFFF) {
            sb.append(hexCharacters[(c >> 28) & 0x0F]);
            sb.append(hexCharacters[(c >> 24) & 0x0F]);
        }
        if(c > 0xFFFF) {
            sb.append(hexCharacters[(c >> 20) & 0x0F]);
            sb.append(hexCharacters[(c >> 16) & 0x0F]);
        }
        if(c > 0xFF) {
            sb.append(hexCharacters[(c >> 12) & 0x0F]);
            sb.append(hexCharacters[(c >>  8) & 0x0F]);
        }
        sb.append(hexCharacters[(c >> 4) & 0x0F]);
        sb.append(hexCharacters[c        & 0x0F]);
        sb.append(";");
        String back = new String(sb);
        return back;
    }



    /**
     * Create HTML representation of a string.
     * @param s Original string.
     * @return  HTML code to show the string.
     */
    public String encodeToHtml(String s) {
        String back = null;
        char c;
        StringBuilder sb = new StringBuilder();
        int l = s.length();
        for(int i = 0; i < l; i++) {
            c = s.charAt(i);
            if((c >= 'a') && (c <= 'z')) {
                sb.append(c);
            } else {
                if((c >= 'A') && (c <= 'Z')) {
                    sb.append(c);
                } else {
                    if(c == '0') {
                        sb.append(c);
                    } else {
                        if((c >= '1') && (c <= '9')) {
                            sb.append(c);
                        } else {
                            switch(c) {
                                case '+':
                                case '-':
                                case '!':
                                case '/':
                                case '(':
                                case ')':
                                case '[':
                                case ']':
                                case '{':
                                case '}':
                                case '=':
                                case '?':
                                case '*':
                                case '~':
                                case '#':
                                case '_':
                                case '.':
                                case ':':
                                case ',':
                                case ';':
                                case '|':
                                {
                                    sb.append(c);
                                } break;
                                case '<': {
                                    sb.append("&lt;");
                                } break;
                                case '>': {
                                    sb.append("&gt;");
                                } break;
                                case ' ': {
                                    sb.append("&nbsp;");
                                } break;
                                default: {
                                    // sb.append(String.format("&#x%04x;", c));
                                    sb.append(htmlEncodedCharacter(c));
                                } break;
                            }
                        }
                    }
                }
            }
        }
        back = new String(sb);
        return back;
    }


}
