/*
Copyright (c) 2009, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.tools;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import javax.swing.TransferHandler;
import javax.swing.JComponent;

/**
 * A transfer handler accepting file names.
 * 
 * @author krause
 *
 */
public abstract class FilenameDropTransferHandler extends TransferHandler
{
  /** class names to check whether or not a class is the String class */
  public static final String[] stringClassNames = {
    "java.lang.String", "String"
  };

  /**
   * While background threads are running we want to disable
   * drag-and-drop
   */
  protected boolean enabled = true;
  
  /**
   * Accept/reject regular files
   */
  protected boolean acceptFiles = true;
  
  /**
   * Accept/reject directories
   */
  protected boolean acceptDirectories = true;
  
  /**
   * Is it sufficient to process one name successfully or
   * do we need to process all names successfully?
   */
  protected boolean mustAcceptAll = false;
  
  /**
   * We can finish data import after first succeeded
   * filename.
   */
  protected boolean finishedAfterFirst = false;
  
  /**
   * Process one filename.
   * 
   * @param fn
   * The file name to process.
   * 
   * @return
   * A flag indicating whether processing was successfully
   * or not.
   */
  protected abstract boolean processFilename(String fn);
  
  /**
   * Check whether or not drag-and-drop is enabled.
   * 
   * @return
   * Check result.
   */
  public synchronized boolean isEnabled()
  {
    return enabled;
  }
  
  /**
   * Enable or disable drag-and-drop.
   * 
   * @param f
   * Enable/disable flag.
   */
  public synchronized void setEnabled(boolean f)
  {
    enabled = f;
  }
  
  /**
   * Check whether or not this TransferHandler accepts
   * file names for regular files.
   * 
   * @return
   * The flag value.
   */
  public synchronized boolean getAcceptFiles()
  {
    return acceptFiles;
  }
  
  /**
   * Set flag for accepting regular file names.
   * 
   * @param f
   * The new flag value.
   */
  public synchronized void setAcceptFiles(boolean f)
  {
    acceptFiles = f;
  }
  
  /**
   * Check whether or not this TransferHandler accepts
   * directory names.
   * 
   * @return
   * The flag value.
   */
  public synchronized boolean getAcceptDirectories()
  {
    return acceptDirectories;
  }
  
  /**
   * Set flag for accepting directory names.
   * 
   * @param f
   * The new flag value.
   */
  public synchronized void setAcceptDirectories(boolean f)
  {
    acceptDirectories = f;
  }
  
  /**
   * Check whether the TransferHander must accept all
   * file names successfully for a successfull dnd
   * operation. Otherwise it is sufficient to process
   * at least on name successfully.
   * 
   * @return
   * The flag value.
   */
  public synchronized boolean getMustAcceptAll()
  {
    return mustAcceptAll;
  }
  
  /**
   * Set flag, whether the TransferHandler must accept
   * all file names successfully for a successfull
   * dnd operation.
   * 
   * @param f
   * The new flag value.
   */
  public synchronized void setMustAcceptAll(boolean f)
  {
    mustAcceptAll = f;
  }
  
  /**
   * Get flag inidicating whether or not we can finish
   * operation after first successful import.
   * 
   * @return
   * The flag.
   */
  public synchronized boolean getFinishedAfterFirst()
  {
    return finishedAfterFirst;
  }
  
  /**
   * Set flag inidicating whether or not we can finish
   * operation after first successful import.
   * 
   * @param b
   * The flag.
   */
  public synchronized void setFinishedAfterFirst(boolean b)
  {
    finishedAfterFirst = b;
  }



  private boolean isStringClassName(String x) {
    boolean back = false;
    if(x != null) {
      if(x.equals(stringClassNames[0])) {
        back = true;
      } else {
        back = x.equals(stringClassNames[1]);
      }
    }
    return back;
  }
    
  /**
   * Extract the file name from a string.
   * 
   * @param s
   * The string containing a file name.
   * 
   * @return
   * The file name.
   */
  private String getFilename(String s)
  {
    String back = null;
    String t = s.trim();
    if(t.startsWith("file://"))
    {
      back = new String(t.substring(7));
    }
    else
    {
      File f = new File(t);
      if(f.isAbsolute() && f.isDirectory()) {
        back = new String(t);
      } f = null;
    } t = null;
    return back;
  }

  /**
   * Check whether or not the contents is usable.
   *
   * @param c
   * The component.
   *
   * @param flavors
   * The collection of contents flavors.
   */
  @Override
  public boolean canImport(JComponent c, DataFlavor[] flavors)
  {
    boolean back = false;
    if(isEnabled()) {
        // file list or string containing file names
      for(int i = 0; i < flavors.length; i++) {
        if(flavors[i].isFlavorJavaFileListType()) {
          back = true;
        } else {
          if(DataFlavor.stringFlavor.equals(flavors[i])) {
            back = true;
          } else {
	    String fvt = flavors[i].getRepresentationClass().getName();
	    if(isStringClassName(fvt)) {
	      back = true;
	    }
	  }
        }
      }
    }
    return back;
  }
  
  /**
   * Test whether or not we can accept a file name drop.
   * 
   * @param f
   * The file name as File.
   * 
   * @return
   * Decision whether to accept or reject the file name.
   */
  private boolean isAcceptableFile(File f)
  {
    boolean back = false;
    if(getAcceptFiles()) {
      if(f.isFile()) {
        back = true;
      }
    }
    if(!back) {
      if(getAcceptDirectories()) {
        if(f.isDirectory()) {
          back = true;
        }
      }
    }
    return back;
  }
  
  /**
   * Import data into component.
   *
   * @param c
   * The component receiving the data.
   *
   * @param t
   * The contents to import.
   *
   * @return
   * Flag to indicate success or error.
   */
  @Override
  @SuppressWarnings("unchecked")
  public boolean importData(JComponent c, Transferable t)
  {
    boolean back = false;
    boolean found = false;
    boolean foundAtLeastOneAcceptableFile = false;
    boolean maa = getMustAcceptAll();
    boolean faf = getFinishedAfterFirst();
    boolean must_continue;
    if(isEnabled()) {
      DataFlavor[] df = t.getTransferDataFlavors();
      for(int i = 0; i < df.length; i++) {
        if(df[i].isFlavorJavaFileListType()) {
          try {
            List<File> fileList = (List<File>)(t.getTransferData(df[i]));
            found = true;
            Iterator<File> it = fileList.iterator();
            File f = null;
            if(maa) {
              back = true;
            } else {
              back = false;
            }
            must_continue = true;
            while(it.hasNext() && must_continue) {
              try {
                f = it.next();
                if(isAcceptableFile(f)) {
                  foundAtLeastOneAcceptableFile = true;
                  if(processFilename(f.getAbsolutePath())) {
                    if(!maa) {
                      back = true;
                    }
                    if(faf) {
                      must_continue = false;
                    }
                  } else {
                    if(maa) {
                      back = false;
                    }
                  }
                } f = null;
              }
              catch(NoSuchElementException e) {
                // should not happen, we asked it.hasNext
              }
            }
            if(maa) {
              if(back) { back = foundAtLeastOneAcceptableFile; }
            } fileList = null; it = null;
          }
          catch(UnsupportedFlavorException e) {
          }
          catch(IOException e) {
          }
        }
      }
      if((!found) && (!back)) {
        for(int i = 0; i < df.length; i++) {
          if(DataFlavor.stringFlavor.equals(df[i])) {
            try {
              String s = (String)(t.getTransferData(DataFlavor.stringFlavor));
              String fn = null;
              StringTokenizer st = new StringTokenizer(s, "\n");
              if(maa) {
                back = true;
              }
              must_continue = true;
              while(st.hasMoreTokens() && must_continue) {
                try {
                  fn = st.nextToken().trim();
                  fn = getFilename(fn);
                  if(fn != null) {
                    File f = new File(fn);
                    if(isAcceptableFile(f)) {
                      foundAtLeastOneAcceptableFile = true;
                      if(processFilename(fn)) {
                        if(!maa) {
                          back = true;
                        }
                        if(faf) {
                          must_continue = false;
                        }
                      } else {
                        if(maa) {
                          back = false;
                        }
                      }
                    } f = null;
                  } fn = null;
                }
                catch(NoSuchElementException e) {
                  // should not happen, we tested using hasMoreTokens()
                }
              }
              if(maa) {
                if(back) { back = foundAtLeastOneAcceptableFile; }
              } s = null; st = null;
            }
            catch(UnsupportedFlavorException e) {
            }
            catch(IOException e) {
            }
          }
        }
      }
      if((!found) && (!back)) {
        for(int i = 0; i < df.length; i++) {
	  String fvt = df[i].getRepresentationClass().getName();
	  if(isStringClassName(fvt)) {
            try {
              String s = (String)(t.getTransferData(df[i]));
              String fn = null;
              StringTokenizer st = new StringTokenizer(s, "\n");
              if(maa) {
                back = true;
              }
              must_continue = true;
              while(st.hasMoreTokens() && must_continue) {
                try {
                  fn = st.nextToken().trim();
                  fn = getFilename(fn);
                  if(fn != null) {
                    File f = new File(fn);
                    if(isAcceptableFile(f)) {
                      foundAtLeastOneAcceptableFile = true;
                      if(processFilename(fn)) {
                        if(!maa) {
                          back = true;
                        }
                        if(faf) {
                          must_continue = false;
                        }
                      } else {
                        if(maa) {
                          back = false;
                        }
                      }
                    } f = null;
                  } fn = null;
                }
                catch(NoSuchElementException e) {
                  // should not happen, we tested using hasMoreTokens()
                }
              }
              if(maa) {
                if(back) { back = foundAtLeastOneAcceptableFile; }
              } s = null; st = null;
            }
            catch(UnsupportedFlavorException e) {
            }
            catch(IOException e) {
            }
	  }
	}
      }
      df = null;
    }
    return back;
  }
}



// vim : set ai sw=2 :

// project dklibsj
// @(#)FilenameDropTransferHandler.java 1.9 06/25/08

