/*
Copyright (c) 2009, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.tools;



import java.io.InputStreamReader;
import java.io.BufferedReader;


/**
 * Class to retrieve platform-dependent settings.
 * In most cases this class will not longer be needed as the system
 * properties contains the information.
 * This class runs the jhelp program.
 * @author krause
 */
public class PlatformInformation {

    /** Command to run. */
    public static final String[] command_to_run = { "jhelp", "platform" };

    /** Keys to retrieve boolean values. */
    public static final String[] kwBoolean = {
        /*   0 */ "filenames.case-sensitive",
        /*   1 */ "filenames.driveletter",
        /*   2 */ "filenames.backslash",
        /*   3 */ "filenames.alldotall",
        /*   4 */ "filenames.dos-style",
        /*   5 */ "filenames.dotfiles",
        /*   6 */ "bigendian",
    };

    /** Boolean values. */
    private boolean[] myBoolean = {
        true,   //  0: file names case-sensitive
        false,  //  1: File name starts with drive letter
        false,  //  2: Backslash used as separator
        false,  //  3: All files must be specified as *.*, not *
        false,  //  4: File names are in DOS 8+3 style
        true,   //  5: Dot files are allowed
        false,  //  6: System is big-endian
    };

    /** String keys. */
    public static final String[] kwString = {
        /*   0 */ "directory.home",
        /*   1 */ "directory.temp",
        /*   2 */ "directory.current",
        /*   3 */ "user.name",
        /*   4 */ "user.gecos",
        /*   5 */ "user.language",
        /*   6 */ "HOME",
        /*   7 */ "LOGNAME",
        /*   8 */ "TEMP",
        /*   9 */ "TMPDIR",
        /*  10 */ "/tmp",
    };

    /** String values. */
    private String[] myString = {
        null,   //  0: Users home directory
        null,   //  1: Temporary directory
        null,   //  2: Current directory
        null,   //  3: Current user name
        null,   //  4: Current user gecos
        null,   //  5: Current user language
        null,   //  6: HOME environment variable
        null,   //  7: LOGNAME environment variable
        null,   //  8: TEMP environment variable
        null,   //  9: TMPDIR environment variable
        null,   // 10: /tmp directory
    };

    /** Int keys. */
    public static final String[] kwInt = {
        /*   0 */ "user.uid",
        /*   1 */ "user.gid",
    };

    /** Int values. */
    private int[] myInt= {
        -1, //  0: User UID
        -1, //  1: User GID
    };

    /**
     * Constructor.
     */
    public PlatformInformation() {
        ProcessBuilder pb = null;
        Process p = null;
        InputStreamReader isr = null;
        BufferedReader br = null;
        String s;
        boolean cc;
        int i;
        try {
            pb = new ProcessBuilder(command_to_run);
            p = pb.start();
            isr = new InputStreamReader(p.getInputStream());
            br = new BufferedReader(isr);
            cc = true;
            while(cc) {
                s = null; cc = false;
                try {
                    s = br.readLine(); cc = true;
                } catch(Exception e) { cc = false; s = null; }
                if(s != null) {
                    processLine(s);
                } else {
                    cc = false;
                }
            }
            try { br.close(); } catch(Exception e) {} br = null;
            try { isr.close(); } catch(Exception e) {} isr = null;
            isr = new InputStreamReader(p.getErrorStream());
            br = new BufferedReader(isr);
            cc = true;
            while(cc) {
                s = null; cc = false;
                try {
                    s = br.readLine(); cc = true;
                } catch(Exception e) { cc = false; s = null; }
                if(s == null) {
                    cc = false;
                }
            }
            try { br.close(); } catch(Exception e) {} br = null;
            try { isr.close(); } catch(Exception e) {} isr = null;
        }
        catch(Exception e) {
        }
        finally {
            if(p != null) { try { p.waitFor(); } catch(Exception e) {}}
        }
        if(myString[5] != null) {
            i = myString[5].indexOf('.');
            if(i > -1) {
                s = myString[5].substring(0, i);
                myString[5] = s;
            }
        }
    }

    /**
     * Process one input line.
     * @param s Input line to process.
     */
    private void processLine(String s) {
        String k;
        String v;
        int i;
        i = s.indexOf(' ');
        if(i >= 0) {
            k = s.substring(0, i);
            v = s.substring(i + 1).trim();
            i = DkTool.getStringArrayIndex(kwBoolean, k, false);
            if(i > -1) {
                myBoolean[i] = DkTool.getBoolean(v, false);
            } else {
                i = DkTool.getStringArrayIndex(kwInt, k, false);
                if(i > -1) {
                    myInt[i] = DkTool.getIntValue(v);
                } else {
                    i = DkTool.getStringArrayIndex(kwString, k, false);
                    if(i > -1) {
                        myString[i] = v;
                    }
                }
            }
        }
    }

    /**
     * Check whether or not file names are case sensitive.
     * @return  Flag: Case sensitive (true) or not (false).
     */
    public boolean getFilenamesCaseSensitive() {
        return myBoolean[0];
    }

    /**
     * Check whether or not file names are allowed to have a drive letter.
     * @return  Flag: Drive letter allowed (true) or not (false).
     */
    public boolean getFilenamesHaveDriveLetter() {
        return myBoolean[1];
    }

    /**
     * Check whether the backslash is the path separator.
     * @return  Flag: backslash (true) or slash (false) as separator.
     */
    public boolean getFilenamesSeparatorBackslash() {
        return myBoolean[2];
    }

    /**
     * Check whether we need all dot all notation to specify all files.
     * @return  Flag: all dot all (true) or all (false).
     */
    public boolean getFilenamesAllDotAll() {
        return myBoolean[3];
    }

    /**
     * Check whether file names must follow the 8+3 DOS convention.
     * @return  Flag: DOS convention required (true) or not (false).
     */
    public boolean getFilenamesDos() {
        return myBoolean[4];
    }

    /**
     * Check whether filenames are allowed to begin with a dot.
     * @return  Flag: Dot files allowed (true) or not (false).
     */
    public boolean getDotFilenames() {
        return myBoolean[5];
    }

    /**
     * Check whether the system is big-endian.
     * @return  Flag: Big-endian (true) or little-endian (false).
     */
    public boolean getBigEndian() {
        return myBoolean[6];
    }

    /**
     * Get the current users home directory.
     * @return  Home directory or null.
     */
    public String getDirectoryHome() {
        String back = myString[0];
        if(back == null) {
            back = myString[6];
        }
        return back;
    }

    /**
     * Get directory for temporary files.
     * @return  Directory name or null.
     */
    public String getDirectoryTemp() {
        String back = myString[1];
        if(back == null) {
            back = myString[8];
            if(back == null) {
                back = myString[9];
                if(back == null) {
                    back = myString[10];
                }
            }
        }
        return back;
    }

    /**
     * Get current directory at start of process.
     * @return  Directory name or null.
     */
    public String getDirectoryCurrent() {
        return myString[2];
    }

    /**
     * Get current user name.
     * @return  User name or null.
     */
    public String getUserName() {
        String back = myString[3];
        if(back == null) {
            back = myString[7];
        }
        return back;
    }

    /**
     * Get current user gecos.
     * @return  Gecos (comment field) or null.
     */
    public String getUserGecos() {
        return myString[4];
    }

    /**
     * Get current users language.
     * @return  Language name (two-letter code) or null.
     */
    public String getUserLanguage() {
        return myString[5];
    }

    /**
     * Get current users UID.
     * @return UID or -1.
     */
    public int getUserUID() {
        return myInt[0];
    }

    /**
     * Get current users GID.
     * @return GID or null.
     */
    public int getUserGID() {
        return myInt[1];
    }

}

