/*
Copyright (c) 2009, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.tools;


import java.io.InputStream;
import java.io.OutputStream;
import java.util.prefs.Preferences;
import java.net.Socket;
import java.net.SocketAddress;
import java.net.InetSocketAddress;



/**	Printer client.
	Objects of this class can send a byte array or a string
	containing the print job data to an LPD server or directly
	to a printer using a direct TCP connection.
*/
public class PrintClient {

  /** Error: Memory allocation problem. */
  public static final int ERROR_MEMORY			=	1;

  /** Error: Unknown protocol, must be "tcp" or "lpd". */
  public static final int ERROR_UNKNOWN_PROTOCOL	=	2;

  /** Error: No destination specified yet. */
  public static final int ERROR_NO_DESTINATION		=	3;

  /** Error: An invalid argument was passed to a method. */
  public static final int ERROR_INVALID_ARGUMENT	=	4;

  /** Error: Failed to create socket. */
  public static final int ERROR_FAILED_TO_CREATE_SOCKET	=	5;

  /** Error: Failed to bind local port. */
  public static final int ERROR_FAILED_LOCAL_PORT	=	6;

  /** Error: Problems with remote address. */
  public static final int ERROR_BAD_REMOTE_ADDRESS	=	7;

  /** Error: Failed to connect to remote address. */
  public static final int ERROR_FAILED_TO_CONNECT	=	8;

  /** Error: Failed to write data. */
  public static final int ERROR_WHILE_WRITING		=	9;

  /** Error: Failed to read response. */
  public static final int ERROR_WHILE_READING		=	10;

  /** Error: Failed to find my own host name. */
  public static final int ERROR_FAILED_TO_FIND_HOSTNAME	=	11;

  /** Error: Request rejected by server. */
  public static final int ERROR_REJECTED_BY_SERVER	=	12;

  /** Protocol: Not yet specified. */
  public static final int PROTOCOL_UNKNOWN		=	0;

  /** Protocol: LPR/LPD. */
  public static final int PROTOCOL_LPR			=	1;

  /** Protocol: Raw TCP. */
  public static final int PROTOCOL_TCP			=	2;

  /** Protocol strings. */
  public static final String[] protocols = {
    /*   0 */	"lpr",
    /*   1 */	"tcp"
  };

  /** Option keywords. */
  public static final String[] optionKeys = {
    /*   0 */	"localport",
    /*   1 */	"response",
    /*   2 */	"ctrl-d",
  };

  /** Preferences root key. */
  public static final String prefRoot = "/printing/lpd";

  /** Key to retrieve job ID. */
  public static final String prefKey = "job-id";

  /** Confirmation byte to send after LPR requests. */
  public static final byte[]	zerobyte = { 0x00 };


  /** Protocol for print job submission. */
  protected int destinationProtocol = PROTOCOL_UNKNOWN;

  /** Host (name or IP) to submit job to. */
  protected String destinationHost = null;

  /** Port number on destination host. */
  protected int destinationPort = 515;

  /** Queue name on destination host. */
  protected String destinationQueue = null;

  /** User name for print job. */
  protected String destinationUser = null;

  /** Error code for last error occured. */
  protected int lastError = 0;

  /** Local port (minimum value). */
  protected int localPortMin = 0;

  /** Local port (maximum value). */
  protected int localPortMax = 0;

  /** Flag: Read response from printer. */
  protected boolean readResponse = false;

  /** Flag: Append CTRL-D to end if not present. */
  protected boolean appendCtrlD = false;

  /** Response from printer. */
  protected String responseString = null;

  /** Tool object. */
  private DkTool dkt = null;



  /**
   * Constructor.
   * @param d   Tool object.
   */
  public PrintClient(DkTool d) {
      dkt = d;
  }



  /**
   * Constructor.
   */
  public PrintClient() {
      this(null);
  }


  
  /**	Retrieve last error code, reset error code.
  	@return	Last error code.
  */
  public int getLastError() {
    int back = lastError;
    lastError = 0;
    return back;
  }


  /**
   * Find job ID for next print job.
   * @return    Job ID.
   */
  private int nextJobId() {
    int back = 1;
    int i;
    try {
      Preferences pun = Preferences.userRoot().node(prefRoot);
      i = pun.getInt(prefKey, 0);
      back = i + 1;
      if(back >= 1000) back = 1;
      pun.putInt(prefKey, back);
    } catch(Exception e) {}
    return back;
  }



  /**	Check whether or not the destination is specified.
  	@return	Flag to indicate success.
  */
  private boolean checkDestination() {
    boolean back = false;
    switch(destinationProtocol) {
      case PROTOCOL_LPR: {
        if(destinationHost != null) {
	  if(destinationHost.length() > 0) {
	    if(destinationPort > 0) {
	      if(destinationQueue != null) {
	        if(destinationQueue.length() > 0) {
		  if(destinationUser != null) {
		    if(destinationUser.length() > 0) {
		      back = true;
		    }
		  }
		}
	      }
	    }
	  }
	}
      } break;
      case PROTOCOL_TCP: {
        if(destinationHost != null) {
	  if(destinationHost.length() > 0) {
	    if(destinationPort > 0) {
	      back = true;
	    }
	  }
	}
      } break;
    }
    return back;
  }



  private byte[] stringToBytes(String s) {
    int lgt;
    int i;
    byte[] back = null;
    lgt = s.length();
    back = new byte[lgt];
    for(i = 0; i < lgt; i++) {
      back[i] = (byte)(s.charAt(i));
    }
    return back;
  }



  private byte[] createRequest(byte c, String s) {
    byte[] back = null;
    int lgt;
    int i;
    
    lgt = s.length();
    back = new byte[lgt+2];
    back[0] = c;
    for(i = 0; i < lgt; i++) {
      back[i+1] = (byte)(s.charAt(i));
    }
    back[i+1] = (byte)'\n';
    return back;
  }



  /**	Send print job using LPR/LPD protocol.
  	@param	printJob	Print job data.
	@return	Flag to indicate success.
  */
  private boolean internalSubmitLpr(byte[] printJob) {
    boolean	back = false;
    boolean	memerr;
    int		jid = nextJobId();
    String	datafilename = null;
    String	ctrlfilename = null;
    String	hostname = DkTool.getHostName();
    String	hn = null;
    int		i1;
    int 	i;
    int		rdbytes;
    StringBuilder	sb = null;
    byte[]	ctrlfiledata = null;
    byte[]	rq1 = null;
    byte[]	rq2 = null;
    byte[]	rq3 = null;
    byte[]	rpb = null;
    boolean	local_port_ok;
    memerr = true;
    
    try {
      if(hostname != null) {
        i1 = hostname.indexOf('.');
        if(i1 > 0) {
          hn = hostname.substring(0, i1);
        } else {
          hn = hostname;
        } 
	datafilename = String.format("dfa%03d%s", jid, hn);
	ctrlfilename = String.format("cfa%03d%s", jid, hn);
	
	
        sb = new StringBuilder();
        sb.append("H");
        sb.append(hn);
        sb.append("\n");
        sb.append("P");
        sb.append(destinationUser);
        sb.append("\n");
        sb.append("l");
        sb.append(datafilename);
        sb.append("\n");
        ctrlfiledata = stringToBytes(new String(sb));
	
        rq1 = createRequest(
	  (byte)0x02,
	  destinationQueue
	); 
        rq2 = createRequest(
          (byte)0x02,
	  String.format("%d %s", ctrlfiledata.length, ctrlfilename)
        ); 
        rq3 = createRequest(
          (byte)0x03,
	  String.format("%d %s", printJob.length, datafilename)
        ); 
	rpb = new byte[1024];
        if((ctrlfiledata != null) && (rq1 != null)
           && (rq2 != null) && (rq3 != null) && (rpb != null))
        {
	  memerr = false; 
	  try {
	    Socket s = new Socket();
	    local_port_ok = true;
	    if((localPortMin > 0) &&(localPortMax >= localPortMin)) {
	      i = localPortMin; 
	      while((i <= localPortMax) && (!local_port_ok)) {
	        try {
		  s.bind(new InetSocketAddress(i));
		  local_port_ok = true;
		} catch(Exception e) {}
	        i++;
	      }
	    } else {	
	    }
	    if(local_port_ok) {
	      try {
	        SocketAddress addr =
		new InetSocketAddress(destinationHost, destinationPort);
		try {
		  s.connect(addr, 5000);	
		  OutputStream os = s.getOutputStream();
		  InputStream is = s.getInputStream();
		  try {
		    os.write(rq1); os.flush();	
		    rdbytes = is.read(rpb);	
		    if(rdbytes > 0) {		
		    if(rpb[0] == 0x00) {	
		      os.write(rq2); os.flush();	
		      rdbytes = is.read(rpb);		
		      if(rdbytes > 0) {
		      if(rpb[0] == 0x00) {	
		        os.write(ctrlfiledata); os.flush();	
			os.write(zerobyte); os.flush();		
			rdbytes = is.read(rpb);		
			if(rdbytes > 0) {
			if(rpb[0] == 0x00) {		
			  os.write(rq3); os.flush();	
			  rdbytes = is.read(rpb);	
			  if(rdbytes > 0) {
			  if(rpb[0] == 0x00) {		
			    os.write(printJob); os.flush();	
			    os.write(zerobyte); os.flush();	
			    rdbytes = is.read(rpb);	
			    if(rdbytes > 0) {
			    if(rpb[0] == 0x00) {	
			      back = true;
			    }
			    }
			  }
			  }
			}
			}
		      }
		      }
		    }
		    }
		    if(!back) {
		      lastError = ERROR_REJECTED_BY_SERVER;
		    }
		  } catch(Exception e) {
		    lastError = ERROR_REJECTED_BY_SERVER;
		  }
		  is.close(); is = null;
		  os.close(); os = null;
		} catch(Exception e) {
		  lastError = ERROR_FAILED_TO_CONNECT;
		}
	      } catch(Exception e) {
	        lastError = ERROR_BAD_REMOTE_ADDRESS;
	      }
	    }
	  } catch(Exception e) {
	    lastError = ERROR_FAILED_TO_CREATE_SOCKET;
	  }
        }
      } else {
        lastError = ERROR_FAILED_TO_FIND_HOSTNAME;
        memerr = false;
      }
    } catch(Exception e) {}
    if(memerr) {
      lastError = ERROR_MEMORY;
    } 
    return back;
  }



  /**	Send print job over raw TCP connction.
  	@param	printJob	Print job data
	@return	Flag to indicate success.
  */
  private boolean internalSubmitTcp(byte[] printJob) {
    boolean back = false;
    boolean local_port_ok = true;
    int	    i;
    
    try {
      Socket s = new Socket();
      if((localPortMin > 0) && (localPortMax >= localPortMin)) {
        local_port_ok = false;
	i = localPortMin;
	while((i <= localPortMax) && (!local_port_ok)) {
	  try {
	    s.bind(new InetSocketAddress(i));
	    local_port_ok = true;
	  } catch(Exception e) {}
	  i++;
	}
      }
      if(local_port_ok) {
        try {
        SocketAddress addr =
	new InetSocketAddress(destinationHost, destinationPort);
	  try {
	    s.connect(addr, 5000);
	    OutputStream os = s.getOutputStream();
	    InputStream is = null;
	    try {
	      os.write(printJob);
	      os.flush();
	      s.shutdownOutput();
	      back = true;
	      if(readResponse) {
	        try {
	          is = s.getInputStream();
	          byte[] responsebytes = new byte[1024];
		  int readbytes = is.read(responsebytes);
		  if(readbytes > 0) {
		    StringBuilder sb = new StringBuilder();
		    for(i = 0; i < readbytes; i++) {
		      sb.append((char)(responsebytes[i]));
		    }
		    responseString = new String(sb);
		  }
		  is.close(); is = null;
	        } catch(Exception e) {
	          lastError = ERROR_WHILE_READING;
		  back = false;
	        }
	      }
	    } catch(Exception e) {
	      lastError = ERROR_WHILE_WRITING;
	    }
	    os.close(); os = null;
	  } catch(Exception e) {
	    lastError = ERROR_FAILED_TO_CONNECT;
	  }
	} catch(Exception e) {
	  lastError = ERROR_BAD_REMOTE_ADDRESS;
	}
      } else {
        lastError = ERROR_FAILED_LOCAL_PORT;
      }
      try { s.close(); } catch(Exception e) {}
    } catch(Exception e) {
      lastError = ERROR_FAILED_TO_CREATE_SOCKET;
    } 
    return back;
  }


  /**	Submit print job.
	@param	printJob	Print job data.
	@return	Flag to indicate success.
  */
  private boolean internalSubmit(byte[] printJob) {
    boolean back = false;
    
    if(checkDestination()) {
      switch(destinationProtocol) {
        case PROTOCOL_LPR: {
	  responseString = null;
	  back = internalSubmitLpr(printJob);
	} break;
	case PROTOCOL_TCP: {
	  responseString = null;
	  back = internalSubmitTcp(printJob);
	} break;
	default: {
	  lastError = ERROR_NO_DESTINATION;
	} break;
      }
    } else {
      lastError = ERROR_NO_DESTINATION;
    } 
    return back;
  }



  /**	Submit print job.
	@param	printJob	Print job data.
	@param	ctrlD		Flag: Must add final CTRL-D if not present.
	@return	Flag to indicate success.
  */
  public boolean submit(byte[] printJob, boolean ctrlD) {
    byte[]	pj = null;
    boolean	back	= false;	/* Function result. */
    byte	b;			/* Test byte. */
    boolean	memerr;			/* Flag: Memory error occured. */
    int		i;
    
    if(printJob != null) {
      if(checkDestination()) {
        if(ctrlD) {
	  b = printJob[printJob.length - 1];
	  if(b == 0x04) {
	    back = internalSubmit(printJob);
	  } else {
	    memerr = true;
	    try {
	      pj = new byte[printJob.length + 1];
	      for(i = 0; i < printJob.length; i++) {
	        pj[i] = printJob[i];
	      }
	      pj[i] = 0x04;
	      memerr = false;
	      back = internalSubmit(pj);
	    } catch(Exception e) {}
	    if(memerr) {
	      lastError = ERROR_MEMORY;
	    }
	  }
	} else {
	  back = internalSubmit(printJob);
	}
      } else {
        lastError = ERROR_NO_DESTINATION;
      }
    } else {
      lastError = ERROR_INVALID_ARGUMENT;
    } 
    return back;
  }



  /**	Submit print job specified as byte array.
  	@param	printJob	Print job data.
	@return	Flag to indicate success.
  */
  public boolean submit(byte[] printJob) {
    boolean back = submit(printJob, appendCtrlD);
    return back;
  }



  /**	Submit print job.
	@param	printJob	Print job data.
	@param	ctrlD		Flag: Must add final CTRL-D if not present.
	@return	Flag to indicate success.
  */
  public boolean submit(String printJob, boolean ctrlD) {
    boolean	back	= false;	/* Return value. */
    byte[]	pj	= null;		/* Print job as byte array. */
    int		lgt;			/* Length of pj. */
    int		i;			/* Running index. */
    byte	b;			/* Test byte to check for CTRL-D. */
    boolean	memerr;			/* Flag: Memory error occured. */
    if(printJob != null) {
      if(checkDestination()) {
        if(ctrlD) {
	  lgt = printJob.length();
	  b = (byte)(printJob.charAt(printJob.length() - 1));
	  memerr = true;
	  try {
	    if(b == 0x04) {
	      pj = new byte[lgt];
	    } else {
	      pj = new byte[lgt + 1];
	    }
	    for(i = 0; i < printJob.length(); i++) {
	      pj[i] = (byte)(printJob.charAt(i));
	    }
	    if(b != 0x04) {
	      pj[i] = 0x04;
	    }
	    memerr = false;
	    back = internalSubmit(pj);
	  } catch(Exception e) {}
	  if(memerr) {
	    lastError = ERROR_MEMORY;
	  }
	} else {
	  memerr = true;
	  try {
	    pj = new byte[printJob.length()];
	    for(i = 0; i < printJob.length(); i++) {
	      pj[i] = (byte)(printJob.charAt(i));
	    }
	    memerr = false;
	    back = internalSubmit(pj);
	  } catch(Exception e) {}
	  if(memerr) {
	    lastError = ERROR_MEMORY;
	  }
	}
      } else {
        lastError = ERROR_NO_DESTINATION;
      }
    } else {
      lastError = ERROR_INVALID_ARGUMENT;
    }
    return back;
  }



  /**	Submit print job.
  	@param	printJob	Print job data as string.
	@return	Flag to indicate success.
  */
  public boolean submit(String printJob) {
    boolean back = submit(printJob, appendCtrlD);
    return back;
  }


  /**	Process additional comma-separated options 
  	response=boolean
	localport=any|min-max.
	@param	options	Options string.
	@return	Flag to indicate success.
  */
  private boolean optionsProcessing(String options) {
    boolean	back = true;
    boolean	cc = true;
    boolean	found_numbers = false;
    String	myText = null;
    String	thisText = null;
    String	nextText = null;
    String	k = null;
    String	v = null;
    String	vv = null;
    String	smin = null;
    String	smax = null;
    int		i1;
    if(options != null) {
      myText = new String(options);
      while(cc && (myText != null)) {
        i1 = myText.indexOf(',');
	if(i1 >= 0) {
	  thisText = myText.substring(0, i1);
	  nextText = myText.substring(i1 + 1);
	} else {
	  thisText = myText; nextText = null; cc = false;
	}
	if(thisText != null) {
	  k = null; v = null;
	  i1 = thisText.indexOf('=');
	  if(i1 > 0) {
	    k = thisText.substring(0, i1);
	    v = thisText.substring(i1 + 1);
	  } else {
	    k = thisText; v = null;
	  }
	  if(k != null) {
	    vv = null;
	    if(v != null) { vv = v.trim(); }
	    switch(DkTool.getStringArrayIndex(optionKeys, k.trim())) {
	      case 0: {	// localport
	        if(vv != null) {
		  if(vv.equals("any") || vv.equals("*")) {
		    localPortMin = localPortMax = 0;
		  } else {
		    i1 = vv.indexOf('-');
		    if(i1 > 0) {
		      smin = vv.substring(0, i1);
		      smax = vv.substring(i1 + 1);
		      found_numbers = false;
			try {
			  localPortMin = Integer.parseInt(smin);
			  localPortMax = Integer.parseInt(smax);
			  found_numbers = true;
			} catch(Exception e) {}
		      if(!found_numbers) {
		        localPortMin = localPortMax = 0;
		      }
		    } else {
		      // ##### ERROR
		    }
		  }
		} else {
		  localPortMin = localPortMax = 0;
		}
	      } break;
	      case 1: {	// response
	        if(vv != null) {
		  if(DkTool.isBoolean(vv)) {
		    readResponse = DkTool.getBoolean(vv, false);
		  } else {
		    readResponse = true;
		  }
		} else {
		  readResponse = true;
		}
	      } break;
	      case 2: {	// ctrl-d
	        if(vv != null) {
		  if(DkTool.isBoolean(vv)) {
		    appendCtrlD = DkTool.getBoolean(vv, false);
		  } else {
		    appendCtrlD = true;
		  }
		} else {
		  appendCtrlD = true;
		} 
	      } break;
	    }
	  }
	}
	thisText = null;
	myText = nextText;
	nextText = null;
      }
    }
    return back;
  }



  /**	Set destination for print jobs submitted later.
  	Destination is in form "tcp:host:port[;options]"
	or "lpd:[user@]host[:port]!queue[;options]".
  	@param	ds	Destination description.
	@return	Flag to indicate success.
  */
  public boolean setDestination(String ds) {
    boolean	back = false;
    boolean	memerr;
    boolean	havePort;
    int		i1;
    String	proto = null;
    String	remainder = null;
    String	rem2 = null;
    String	hostport = null;
    String	rem3 = null;
    String	options = null;
    String	rem4 = null;
    
    if(ds != null) {
      i1 = ds.indexOf(':');
      if(i1 > 0) {
        memerr = true;
        try {
          proto = ds.substring(0, i1);
	  remainder = ds.substring(i1+1);
	  switch(DkTool.getStringArrayIndex(protocols, proto)) {
	    case 0: {
	      destinationUser = null;
	      destinationProtocol = PROTOCOL_LPR;
	      destinationPort = 515;
	      localPortMin = 721; localPortMax = 731;
	      i1 = remainder.indexOf('@');
	      if(i1 > 0) {
	        destinationUser = remainder.substring(0, i1);
		rem2 = remainder.substring(i1 + 1);
	      } else {
		destinationUser = null;
                if(dkt != null) {
                    destinationUser = dkt.getUserName();
                }
		rem2 = remainder;
	      }
	      if(destinationUser != null) {
		i1 = rem2.indexOf('!');
		if(i1 > 0) {
		  hostport = rem2.substring(0, i1);
		  rem3 = rem2.substring(i1 + 1);
		  i1 = hostport.indexOf(':');
		  havePort = false;
		  if(i1 >= 0) {
		    destinationHost = hostport.substring(0, i1);
		    rem4 = hostport.substring(i1 + 1);
		    try {
		      destinationPort = Integer.parseInt(rem4);
		      havePort = true;
		    } catch(Exception e) {
		      lastError = ERROR_INVALID_ARGUMENT; memerr = false;
		    }
		  } else {
		    destinationHost = hostport; havePort = true;
		  }
		  if(havePort) {
		    i1 = rem3.indexOf(';');
		    if(i1 > 0) {
		      destinationQueue = rem3.substring(0, i1);
		      options = rem3.substring(i1 + 1);
		    } else {
		      destinationQueue = rem3;
		    }
		    back = true;
		    if(options != null) {
		      if(!optionsProcessing(options)) {
		        back = false;
		      }
		    }
		  }
		} else {
		  lastError = ERROR_INVALID_ARGUMENT; memerr = false;
		}
	      } else {
	        lastError = ERROR_INVALID_ARGUMENT; memerr = false;
	      }
	    } break;
	    case 1: {
	      destinationProtocol = PROTOCOL_TCP;
	      localPortMin = 0; localPortMax = 0;
	      i1 = remainder.indexOf(':');
	      if(i1 > 0) {
	        destinationHost = remainder.substring(0, i1);
		rem2 = remainder.substring(i1 + 1);
		String portno = null;
		boolean	havePortno = false;
		i1 = rem2.indexOf(';');
		if(i1 > 0) {
		  portno = rem2.substring(0, i1);
		  options = rem2.substring(i1 + 1);
		} else {
		  portno = rem2;
		}
		try {
		  destinationPort = Integer.parseInt(portno);
		  havePortno = true;
		} catch(Exception e) {
		  lastError = ERROR_INVALID_ARGUMENT; memerr = false;
		}
		if(havePortno) {
		  back = true;
		  if(options != null) {
		    if(!optionsProcessing(options)) {
		      back = false;
		    }
		  }
		}
	      } else {
	        lastError = ERROR_INVALID_ARGUMENT; memerr = false;
	      }
	    } break;
	    default: {
	      memerr = false; lastError = ERROR_INVALID_ARGUMENT;
	    } break;
	  }
	} catch(Exception e) {}
	if(memerr) {
	  lastError = ERROR_MEMORY;
	}
      } else {
        lastError = ERROR_INVALID_ARGUMENT;
      }
    } else {
      lastError = ERROR_INVALID_ARGUMENT;
    }
    if(!back) {
      destinationProtocol = PROTOCOL_UNKNOWN;
      destinationHost = null;
      destinationPort = 515;
      destinationQueue = null;
      destinationUser = null;
      localPortMin = localPortMax = 0;
      readResponse = appendCtrlD = false;
    } 
    return back;
  }



  public String getResponse() {
    String back = null;
    if(responseString != null) {
      back = new String(responseString);
    }
    return back;
  }

  private String getSystemUser() {
    String back = null;
    try {
      back = System.getenv("LOGNAME");
    } catch(Exception e) { back = null; }
    if(back == null) {
      try {
        back = System.getenv("USERNAME");
      } catch(Exception e) { back = null; }
    }
    return back;
  }

}
