/*
Copyright (c) 2001-2008, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.jtools;

import java.util.ArrayList;
import java.util.Iterator;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.File;



/**
 * This class executes one or multiple commands and
 * saves output to show it in a LogWindow.
 * 
 * @author krause
 *
 */
public class LogContentsForCommand extends LogContents
{
  /**
   * Some text fragments needed to build the log text.
   */
  public static final String[] msgTexts = {
    "Command: ",
    "Output:",
    "Diagnostic output:",
    "\"",
    "Exit code: ",
    "",
    " ",
    "ERROR: Failed to run the command shown in the line below!"
  };

  /** Directory, in which the command must be executed */
  private String directoryForCommand = null;
  
  /**
   * The command line (command and arguments).
   */
  private ArrayList<String> argv = null;

  
  /**
   * Flag to indicate whether or not a command
   * (at least one) was executed.
   */
  private boolean alreadyExecutedCommand = false;
  
  /**
   * Default constructor.
   *
   */
  public LogContentsForCommand()
  {
    argv = new ArrayList<String>();
  }
  
  /**
   * Prepare to start a new command.
   * This replaces a new command line string collection
   * by an empty new one.
   *
   */
  public void newCommand()
  {
    argv = new ArrayList<String>();
    directoryForCommand = null;
  }




  /**
   * Set directory in which the command is executed.
   *
   * @param d
   * Directory name.
   */
  public void setDirectory(String d) {
    directoryForCommand = d;
  }



  /**
   * Add a string to the command line string collection
   * (command first, arguments follow).
   * 
   * @param s
   * The string to add.
   */
  public void addArgv(String s)
  {
    argv.add(s);
  }
  
  /**
   * Run the command and save output, build a log
   * text to be shown in a LogWindow.
   *
   */
  public void run()
  {
    String[] cmd = new String[argv.toArray().length];
    for(int i = 0; i < argv.toArray().length; i++) {
      cmd[i] = (String)(argv.toArray()[i]);
    }
    if(cmd.length > 0) {
      if(alreadyExecutedCommand) { add(msgTexts[5]); }
      StringBuilder sb = new StringBuilder();
      for(int i = 0; i < cmd.length; i++) {
        if(i > 0) {
          sb.append(msgTexts[6]);
        }
        sb.append(cmd[i]);
      }
      String cmdLine = new String(sb); sb = null;
      try {
        boolean didShowCmd = false;
        boolean didShowStd = false;
        boolean didShowErr = false;
        // Process p = Runtime.getRuntime().exec(cmd);
	ProcessBuilder pbuild = new ProcessBuilder(argv);
	if(directoryForCommand != null) {
	  pbuild.directory(new File(directoryForCommand));
	}
	Process p = pbuild.start();
        InputStreamReader isr = new InputStreamReader(p.getInputStream());
        BufferedReader in;
        in = new BufferedReader(isr);
        for(String s; (s = in.readLine()) != null; ) {
          if(!didShowCmd) {
            StringBuilder sb2 = new StringBuilder(msgTexts[0]);
            sb2.append(
	      msgTexts[3]); sb2.append(cmdLine); sb2.append(msgTexts[3]
	    );
            add(new String(sb2)); sb2 = null;
            didShowCmd = true;
          }
          if(!didShowStd) {
            add(msgTexts[1]);
            didShowStd = true;
          }
          add(s.trim());
        }
        in.close(); in = null; isr = null;
        isr = new InputStreamReader(p.getErrorStream());
        in = new BufferedReader(isr);
        for(String s; (s = in.readLine()) != null; ) {
          if(!didShowCmd) {
            StringBuilder sb2 = new StringBuilder(msgTexts[0]);
            sb2.append(
	      msgTexts[3]); sb2.append(cmdLine); sb2.append(msgTexts[3]
	    );
            add(new String(sb2)); sb2 = null;
            didShowCmd = true;
          }
          if(!didShowErr) {
            add(msgTexts[1]);
            didShowErr = true;
          }
          add(s.trim());
        } in.close(); in = null; isr = null;
        p.waitFor();
        int exitCode = p.exitValue();
        if(exitCode != 0) {
          success = false;
        }
        if(!didShowCmd) {
          StringBuilder sb2 = new StringBuilder(msgTexts[0]);
          sb2.append(
	    msgTexts[3]); sb2.append(cmdLine); sb2.append(msgTexts[3]
	  );
          add(new String(sb2)); sb2 = null;
          didShowCmd = true;
        }
        add(msgTexts[4] + exitCode);
      }
      catch(InterruptedException e) {
        add(msgTexts[7]); add(cmdLine); success = false;
      }
      catch(IOException e) {
        add(msgTexts[7]); add(cmdLine); success = false;
      }
      alreadyExecutedCommand = true;
    }
  }
  
  /**
   * Create an iterator to traverse output line by line.
   * 
   * @return
   * The iterator to use for traversal.
   */
  public Iterator<String> getIterator()
  {
    Iterator<String> back = textLines.iterator();
    return back;
  }
  
}

// vim: set ai sw=2 :


// project dklibsj
// @(#)LogContentsForCommand.java 1.9 02/02/09

