/*
Copyright (c) 2008, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.jtools;

import java.io.InputStreamReader;
import java.io.BufferedReader;


/**
 *
 * @author krause
 */
public class PlatformInformation {

  /** The keys issued by "jhelp platform" */
  public static final String[] jhelpKeys = {
    "filenames.case-sensitive ",	//  0
    "filenames.driveletter ",		//  1
    "filenames.backslash ",		//  2
    "filenames.alldotall ",		//  3
    "filenames.dos-style ",		//  4
    "filenames.dotfiles ",		//  5
    "bigendian ",			//  6
    "directory.temp ",			//  7
    "directory.current ",		//  8
    "directory.home ",			//  9
    "user.name ",			// 10
    "user.uid ",			// 11
    "user.gid ",			// 12
    "user.gecos ",			// 13
  };

  /** The command to run from the init() method */
  public static final String[] jhelpCommand = {
    "jhelp", "platform"
  };


  /** The boolean values array. */
  protected boolean[] bv = {
    true,	// 0: file names case-sensitive
    false,	// 1: file names contain a drive letter
    false,	// 2: file names use backslashes
    false,	// 3: the pattern for all files is *.* instead of *
    false,	// 4: DOS style file names 8+3
    true,	// 5: files with leading .
    false,	// 6: integers are stored bigendian
  };

  /** The string values array. */
  protected String[] sv = {
    "",		// 0: user name
    "",		// 1: home directory
    "",		// 2: user gecos
    "",		// 3: temporary directory
    ".",	// 4: current directory
    "",		// 5: empty default string
  };

  /** The integer values array. */
  private int[] iv = {
    -1,		// 0: UID
    -1,		// 1: GID
  };



  /**
   * Create a new data set using default values.
   *
   * Run the init() method to obtain useful data.
   */
  public PlatformInformation() {
  }



  /**
   * Get one boolean value from the array.
   *
   * @param i
   * The index.
   *
   * @return
   * The value.
   */
  private boolean getBoolean(int i) {
    boolean back = false;
    if((i >= 0) && (i < bv.length)) { back = bv[i]; }
    return back;
  }



  /**
   * Get one string value from the array.
   *
   * @param i
   * The index.
   *
   * @return
   * The value.
   */
  private String getString(int i) {
    String back = sv[5];
    if((i >= 0) && (i < sv.length)) { back = sv[i]; }
    return back;
  }



  /**
   * Get one entry from the integers array.
   *
   * @param i
   * The index.
   *
   * @return
   * The value.
   */
  private int getInt(int i) {
    int back = 0;
    if((i >= 0) && (i < iv.length)) { back = iv[i]; }
    return back;
  }
  


  /**
   * Clone a PlatformInformation data set from an existing
   * data set.
   *
   * @param ori
   * The original data set.
   */
  public void cloneFrom(PlatformInformation ori) {
    int i;
    for(i = 0; i < bv.length; i++) {
      bv[i] = ori.getBoolean(i);
    }
    for(i = 0; i < sv.length; i++) {
      sv[i] = ori.getString(i);
    }
    for(i = 0; i < iv.length; i++) {
      iv[i] = ori.getInt(i);
    }
  }



  /**
   * Find out whether file names are case-sensitive on the system.
   *
   * @return
   * Flag indicating case-sensitive file names.
   */
  public boolean isCaseSensitive() {
    return getBoolean(0);
  }



  /**
   * Find out whether absolute path names have a drive letter at
   * start.
   *
   * @return
   * Flag to indicate the use of drive letters (like C:).
   */
  public boolean isUsingDriveLetter() {
    return getBoolean(1);
  }



  /**
   * Find out whether the backslash is used for file name separations.
   *
   * @return
   * Flag to indicate that backslashes are used (not slashes).
   */
  public boolean isUsingBackslash() {
    return getBoolean(2);
  }



  /**
   * Find out whether the system needs pattern "*.*" to access all files
   * rather than "*"
   *
   * @return
   * Flag to indicate that "*.*" is needed.
   */
  public boolean isUsingAllDotAll() {
    return getBoolean(3);
  }



  /**
   * Find out whether the system is restricted to 8+3 file names.
   *
   * @return
   * Flag to indicate the restriction to 8+3 file names.
   */
  public boolean isUsingDosStyleFileNames() {
    return getBoolean(4);
  }



  /**
   * Find out whether file names are allowed to have a leading dot.
   *
   * @return
   * Flag to indicate that ".*" file names are allowed.
   */
  public boolean isLeadingDotAvailable() {
    return getBoolean(5);
  }



  /**
   * Find out, whether the system stores integer data big-endian.
   *
   * @return
   * The big-endian flag.
   */
  public boolean isBigEndian() {
    return getBoolean(6);
  }



  /**
   * Get the users login name.
   *
   * @return
   * The login name.
   */
  public String getUserName() {
    return getString(0);
  }



  /**
   * Get the users home directory.
   *
   * @return
   * The home directory.
   */
  public String getHomeDirectory() {
    return getString(1);
  }



  /**
   * Get the users gecos field.
   *
   * @return
   * The user comment.
   */
  public String getUserComment() {
    return getString(2);
  }



  /**
   * Get the temporary directory.
   *
   * @return
   * The temporary directory.
   */
  public String getTempDirectory() {
    return getString(3);
  }



  /**
   * Get the original current working directory.
   *
   * @return
   * The current working directory when the application started.
   */
  public String getOriginalCurrentDirectory() {
    return getString(4);
  }



  /**
   * Get the users user ID
   *
   * @return
   * The UID.
   */
  public int getUid() {
    return getInt(0);
  }



  /**
   * Get the users group ID
   *
   * @return
   * The GID:
   */
  public int getGid() {
    return getInt(1);
  }



  /**
   * Retrieve an integer value from one text line.
   *
   * @param s
   * The line to process.
   *
   * @param ind
   * The index in the jhelpKeys array.
   *
   * @param def
   * A default value to use if there is an exception.
   *
   * @return
   * The integer value.
   */
  private int retrieveInt(String s, int ind, int def) {
    int back = def;
    try {
      String ss = s.substring(jhelpKeys[ind].length());
      int i = Integer.parseInt(ss);
      back = i;
    } catch(Exception exc) {}
    return back;
  }



  /**
   * Process one text line retrieved from "jhelp platform".
   *
   * @param s
   * The line to process.
   *
   * @return
   * Success (line contained a good keyword).
   */
  private boolean addLine(String s) {
    boolean back = true;
    switch(StringTool.getStartArrayIndex(jhelpKeys, s)) {
      case 0: {		// filenames case sensitive
	bv[0] = StringTool.getBoolean(s.substring(jhelpKeys[0].length()), bv[0]);
      } break;
      case 1: {		// drive letter
        bv[1] = StringTool.getBoolean(s.substring(jhelpKeys[1].length()), bv[1]);
      } break;
      case 2: {		// backslash
        bv[2] = StringTool.getBoolean(s.substring(jhelpKeys[2].length()), bv[2]);
      } break;
      case 3: {		// *.*
        bv[3] = StringTool.getBoolean(s.substring(jhelpKeys[3].length()), bv[3]);
      } break;
      case 4: {		// 8+3
        bv[4] = StringTool.getBoolean(s.substring(jhelpKeys[4].length()), bv[4]);
      } break;
      case 5: {		// .*
        bv[5] = StringTool.getBoolean(s.substring(jhelpKeys[5].length()), bv[5]);
      } break;
      case 6: {		// bigendian
        bv[6] = StringTool.getBoolean(s.substring(jhelpKeys[6].length()), bv[6]);
      } break;
      case 7: {		// temp dir
        sv[3] = s.substring(jhelpKeys[7].length());
      } break;
      case 8: {		// current dir
        sv[4] = s.substring(jhelpKeys[8].length());
      } break;
      case 9: {		// home dir
        sv[1] = s.substring(jhelpKeys[9].length());
      } break;
      case 10: {	// user name
        sv[0] = s.substring(jhelpKeys[10].length());
      } break;
      case 11: {	// user UID
        iv[0] = retrieveInt(s, 11, iv[0]);
      } break;
      case 12: {	// user GID
        iv[1] = retrieveInt(s, 12, iv[1]);
      } break;
      case 13: {	// gecos
        sv[2] = s.substring(jhelpKeys[13].length());
      } break;
      default: {
        back = false;
      } break;
    }
    return back;
  }
 


  /**
   * Initialize, run "jhelp platform" to retrieve settings.
   */
  public boolean init() {
    boolean back = false;
    boolean good_line_found = false;
    boolean bad_line_found = false;
    try {
      ProcessBuilder pb = new ProcessBuilder(jhelpCommand[0], jhelpCommand[1]);
      Process p = pb.start();
      InputStreamReader isr = new InputStreamReader(p.getInputStream());
      BufferedReader in = new BufferedReader(isr);
      for(String s; (s = in.readLine()) != null; ) {
        if(addLine(s.trim())) {
	  good_line_found = true;
	} else {
	  bad_line_found = true;
	}
      }
      in.close(); in = null; isr = null;
      isr = new InputStreamReader(p.getErrorStream());
      in = new BufferedReader(isr);
      for(String s; (in.readLine()) != null; ) { }
      in.close(); in = null; isr = null;
      p.waitFor();
    }
    catch(Exception exc) {}
    if(good_line_found) {
      back = true;
    }
    if(bad_line_found) {
      back = false;
    }
    return back;
  }


}



// vim : set ai sw=2 :

// project dklibsj
// @(#)PlatformInformation.java 1.3 07/01/08

