/*
Copyright (c) 2001-2009, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package dirk_krause.jtools;

import java.awt.Color;

import java.util.ArrayList;
import java.util.StringTokenizer;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;

import java.net.URL;

/**
 * The static methods of the StringTool class can
 * be used to convert strings to integer, colors and
 * booleans.
 * 
 * @author krause
 *
 */
public class StringTool {
  /**
   * Strings which can be converted into booleans.
   */
  public static final String[] theBooleanStrings = {
    "y", "yes", "on", "true", "ok",
    "n", "no", "off", "false"
  };
  
  /**
   * Parse string for integer value.
   * The string may start with "0x" or "0" to indicate
   * hexadecimal or octal numbers.
   * 
   * @param s
   * The string to parse.
   * 
   * @param d
   * A default value.
   * 
   * @return
   * The result.
   */
  public static int getIntValue(String s, int d) {
    int back = 0;
    if(s != null) {
      if(s.startsWith("0x")) {
        try {
          back = Integer.parseInt(s.substring(2), 16);
        }
        catch(NumberFormatException e) {
          back = d;
        }
      } else {
        if(s.startsWith("0")) {
          try {
            back = Integer.parseInt(s, 8);
          }
          catch(NumberFormatException e) {
            back = d;
          }
        } else {
          try {
            back = Integer.parseInt(s);
          }
          catch(NumberFormatException e) {
            back = d;
          }
        }
      }
    }
    return back;
  }
  
  /**
   * Parse string for integer value.
   * The string may start with "0x" or "0" to indicate
   * hexadecimal or octal numbers.
   * 
   * @param s
   * The string to parse.
   * 
   * @return
   * The result.
   */
  public static int getIntValue(String s) {
    int back = getIntValue(s, 0);
    return back;
  }
  
  /**
   * Convert string into a color.
   * 
   * @param s
   * A string like "128,0,0".
   *
   * @param dr
   * Default for red.
   *
   * @param dg
   * Default for green.
   *
   * @param db
   * Default for blue.
   * 
   * @return
   * The color.
   */
  public static Color getColor(String s, int dr, int dg, int db) {
    int r = dr, g = dg, b = db;
    if(s != null) {
      StringTokenizer st = new StringTokenizer(s, ",");
      if(st.hasMoreTokens()) {
        String rs = st.nextToken();
        r = getIntValue(rs.trim(), 255);
        if(st.hasMoreTokens()) {
          String gs = st.nextToken();
          g = getIntValue(gs.trim(), 255);
          if(st.hasMoreTokens()) {
            String bs = st.nextToken();
            b = getIntValue(bs.trim(), 255);
          }
        } rs = null;
      } st = null;
    }
    return(new Color(r, g, b));
  }



  /**
   * Convert a String into a color.
   *
   * @param s
   * The string to analyze.
   *
   * @return
   * The color.
   */
  public static Color getColor(String s) {
    return getColor(s, 255, 255, 255);
  }
  
  /**
   * Find index of string s in array a.
   * 
   * @param a
   * An array of strings.
   * 
   * @param s
   * The string to search for.
   * 
   * @return
   * On success: the index of s in a (0, 1, ...)
   * On error: -1.
   */
  public static int getArrayIndex(String[] a, String s) {
    int back = -1;
    int i = 0;
    while((i < a.length) && (back == -1)) {
      if(s.equals(a[i])) {
        back = i; i = a.length;
      } else {
        i++;
      }
    }
    return back;
  }

  /**
   * Check whether or not the string s starts with one of
   * the strings from the array a.
   *
   * @param a
   * An array of strings, one of the strings is probably
   * at the start of string s.
   *
   * @param s
   * The string to check.
   *
   * @return
   * The index of that string in a which was found at the
   * start of s. On error -- no matching string found: -1.
   */
  public static int getStartArrayIndex(String[] a, String s) {
    int back = -1;
    int i = 0;
    while((i < a.length) && (back == -1)) {
      if(s.startsWith(a[i])) {
        back = i; i = a.length;
      } else {
        i++;
      }
    }
    return back;
  }
  
  /**
   * Check whether or not a string represents a boolean.
   * 
   * @param s
   * The string to check.
   * 
   * @return
   * A flag whether or not the string represents
   * a boolean.
   */
  public static boolean isBoolean(String s) {
    boolean back = false;
    if(getArrayIndex(theBooleanStrings, s) > -1) {
      back = true;
    }
    return back;
  }
  
  /**
   * Convert a string into a boolean.
   * 
   * @param s
   * The source string.
   * 
   * @param d
   * A default value, returned if the string does
   * not represent a boolean.
   * 
   * @return
   * The boolean value.
   */
  public static boolean getBoolean(String s, boolean d) {
    boolean back = d;
    int i = getArrayIndex(theBooleanStrings, s);
    if(i > -1) {
      if(i < 5) back = true;
      else back = false;
    }
    return back;
  }

  private static final String psAllowedCharacters =
  "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789+-*/?.,;";

  public static String encodeForPostscript(String s) {
      String back;
      StringBuilder res = new StringBuilder();
      int i;
      char c;
      int ic;
      String x;
      for(i = 0; i < s.length(); i++) {
          c = s.charAt(i);
          if(psAllowedCharacters.indexOf((int)c) >= 0) {
              res.append(c);
          } else {
              switch(c) {
                  case '\n': { res.append("\\n"); } break;
                  case '\t': { res.append("\\t"); } break;
                  case '\f': { res.append("\\f"); } break;
                  case '\r': { res.append("\\r"); } break;
                  case '\b': { res.append("\\b"); } break;
                  case '\\': { res.append("\\\\"); } break;
                  default: {
                      ic = (int)c;
                      if((ic >= 0) && (ic <= 255)) {
                          x = Integer.toOctalString(ic);
                          switch(x.length()) {
                              case 0: { res.append("\\000"); } break;
                              case 1: {
                                  res.append("\\00"); res.append(x);
                              } break;
                              case 2: {
                                  res.append("\\0"); res.append(x);
                              } break;
                              case 3: {
                                  res.append("\\"); res.append(x);
                              } break;
                          }
                      }
                  } break;
              }
          }
      }
      back = new String(res);
      return back;
  }

  static private String[] getArrayFromInputStream(InputStream is, String csn) {
    String[] back = null;
    InputStreamReader isr = null;
    BufferedReader br = null;
    ArrayList<String> al = new ArrayList<String>();
    try {
	if(csn != null) { isr = new InputStreamReader(is, csn); }
	else { isr = new InputStreamReader(is); }
	br = new BufferedReader(isr);
	String lineBuffer;
	boolean cc;
	cc = true;
	while(cc) {
	    lineBuffer = null; cc = false;
	    try {
	        lineBuffer = br.readLine();
	    }
	    catch(Exception e) {
	        lineBuffer = null;
	    }
	    if(lineBuffer != null) {
	        cc = true;
		if(!(lineBuffer.startsWith("#"))) {
		    lineBuffer = lineBuffer.replaceAll("\n", "");
		    lineBuffer = lineBuffer.replaceAll("\\#", "#");
		    lineBuffer = lineBuffer.replaceAll("\\n", "\n");
		    lineBuffer = lineBuffer.replaceAll("\\t", "\t");
		    lineBuffer = lineBuffer.replaceAll("\\\\", "\\");
		    al.add(lineBuffer);
		}
	    }
	}
        Object[] objarr = al.toArray();
        back = new String[objarr.length];
        for(int i = 0; i < objarr.length; i++) {
            back[i] = new String((String)(objarr[i]));
        }
    }
    catch(Exception e) {
        back = null;
    }
    finally {
        try { br.close(); } catch(Exception e) {}
	try { isr.close(); } catch(Exception e) {}
    }
    return back;
  }

  /**
   * Read array of strings from a resource.
   * @param	o	Object originating the request.
   * @param	rn	Resource name (file name in JAR file or URL).
   * @param	csn	Character set name ("UTF-8" or "ISO-8859-1").
   * @return	String array on success, null on error.
   */
  public static String[] getArrayFromNamedResource(
      Object o, String rn, String csn
  )
  {
      String[] back = null;
      InputStream is = null;
      if(rn.startsWith("http://")) {
          try {
	      is = (new URL(rn)).openStream();
	      back = getArrayFromInputStream(is, csn);
	  }
	  catch(Exception e) {
	      back = null;
	  }
	  finally {
	      try { is.close(); } catch(Exception e) {}
	  }
      } else {
          try {
	      if(rn.startsWith("/")) {
                  is = o.getClass().getResourceAsStream(rn);
	      } else {
	          is = o.getClass().getResourceAsStream("/" + rn);
	      }
	      back = getArrayFromInputStream(is, csn);
	  }
	  catch(Exception e) {
	      back = null;
	  }
	  finally {
	      try { is.close(); } catch(Exception e) {}
	  }
      }
      return back;
  }


  /**
   * Read array of strings from a resource.
   * @param	o	Object originating the request.
   * @param	rn	Resource name (file name in JAR file or URL).
   * @return	String array on success, null on error.
   */
  public static String[] getArrayFromNamedResource(
      Object o, String rn
  )
  {
      String[] back = null;
      back = getArrayFromNamedResource(o, rn, null);
      return  back;
  }


}



// project dklibsj
// @(#)StringTool.java 1.14 12/07/09

